import copy
from typing import List, Set, TYPE_CHECKING
if TYPE_CHECKING:
    from flitsr.args import Args
from flitsr.spectrum import Spectrum
from flitsr.errors import warning
from flitsr.ranking import Ranking, Tiebrk, set_orig, unset_orig
from flitsr.advanced.ranker import Ranker
from flitsr.advanced.sbfl import SBFL
from flitsr.advanced.attributes import existing, choices, print_name
from flitsr import advanced


class Flitsr(Ranker):
    """
    Run the main FLITSR algorithm over the spectrum to produce ranked lists.
    """
    @existing('tiebrk')
    @existing('args')
    @choices('internal_ranking', ['auto', 'conf', 'original', 'reverse',
                                  'flitsr'])
    def __init__(self, args: 'Args', internal_ranking: str = 'auto',
                 tiebrk: Tiebrk = Tiebrk.ORIG):
        self.tiebrk = tiebrk
        self.args = args
        self.order_method = internal_ranking

    def remove_faulty_elements(self, spectrum: Spectrum,
                               tests_removed: Set[Spectrum.Test],
                               faulty: List[Spectrum.Group]):
        """Removes all tests that execute an 'actually' faulty element"""
        toRemove = []
        for test in tests_removed:
            for f in faulty:
                if (spectrum[test][f] is True):
                    toRemove.append(test)
                    break
        tests_removed.difference_update(toRemove)

    def run_metric(self, spectrum: Spectrum, formula: str):
        if (hasattr(advanced.RankerType, formula.upper())):
            ranker_args = self.args.get_arg_group(formula)
            ranker = advanced.RankerType[formula.upper()].value(**ranker_args)
            ranking = ranker.rank(spectrum, formula)
        else:
            sbfl_args = self.args.get_arg_group('SBFL')
            ranking = SBFL(**sbfl_args).rank(spectrum, formula)
        return ranking

    def flitsr(self, spectrum: Spectrum, formula: str) -> List[Spectrum.Group]:
        """
        Executes the recursive flitsr algorithm to identify faulty elements
        """
        if (spectrum.tf == 0):
            return []
        ranking = self.run_metric(spectrum, formula)
        r_iter = iter(ranking)
        group = next(r_iter).group
        tests_removed = spectrum.get_tests(group, only_failing=True,
                                           remove=True)
        while (len(tests_removed) == 0):  # sanity check
            if ((s2 := next(r_iter, None)) is None):
                count_non_removed = len(spectrum.failing())
                warning("flitsr found", count_non_removed,
                        "failing test(s) that it could not explain")
                return []
            # continue trying the next element if available
            group = s2.group
            tests_removed = spectrum.get_tests(group, only_failing=True,
                                               remove=True)
        faulty = self.flitsr(spectrum, formula)
        self.remove_faulty_elements(spectrum, tests_removed, faulty)
        if (len(tests_removed) > 0):
            faulty.append(group)
        return faulty

    def get_inverse_confidence_scores(self, spectrum: Spectrum,
                                      basis: List[Spectrum.Group]) -> List[int]:
        """
        Using the given basis and spectrum, calculates the confidence scores
        for each basis element. The confidence scores returned are the inverse
        of the actual confidence scores.
        """
        confs: List[int] = []
        for group in basis:
            ts = list(spectrum.get_tests(group, only_failing=True))
            possibles: Set[Spectrum.Group] = set()
            possibles.update(spectrum.get_executed_groups(ts[0]))
            for test in ts[1:]:
                possibles.intersection_update(spectrum.get_executed_groups(test))
            confs.append(len(possibles))
        return confs

    def flitsr_ordering(self, spectrum: Spectrum, basis: List[Spectrum.Group],
                        ranking: Ranking,
                        flitsr_order='auto') -> List[Spectrum.Group]:
        """
        Order the given flitsr basis using the specified ordering strategy.
        Strategies include 'auto', 'conf', 'original', 'reverse', and 'flitsr'.
        'flitsr' preserves the current ordering of the basis, 'reverse'
        reverses this order (i.e. order elements were identified by flitsr),
        'original' orders elements by their position in the original ranking
        (given by ranking). 'conf' gets the confidence scores for each element
        and orders by these. 'auto' also gets the confidence scores for each
        element and uses them to pick which strategy to use (either 'flitsr',
        'original' or 'conf').
        """
        if (len(basis) == 0):
            return basis
        inv_confs = []
        # check if internal ranking order needs to be determined
        if (flitsr_order in ['auto', 'conf']):
            inv_confs = self.get_inverse_confidence_scores(spectrum, basis)
        if (flitsr_order == 'auto'):
            if (all(c > 3 for c in inv_confs)):
                flitsr_order = 'original'
            elif (all(c <= 3 for c in inv_confs)):
                flitsr_order = 'flitsr'
            else:
                flitsr_order = 'conf'
            # check for big groups
            big, small = [], []
            for group in basis:
                if (len(group.get_elements()) > 5):
                    big.append(group)
                else:
                    small.append(group)
            if (len(big) != 0 and len(small) != 0):
                return self.flitsr_ordering(spectrum, small, ranking,
                                            flitsr_order) + \
                    self.flitsr_ordering(spectrum, big, ranking, flitsr_order)
        # reorder basis
        if (flitsr_order == 'flitsr'):
            ordered_basis = basis
        elif (flitsr_order == 'reverse'):
            ordered_basis = list(reversed(basis))
        elif (flitsr_order == 'original'):
            ordered_basis = [x.group for x in ranking if x.group in basis]
            # add any missing elements
            if (len(ordered_basis) < len(basis)):
                ordered_basis.extend([e for e in basis if e not in ordered_basis])
        elif (flitsr_order == 'conf'):
            ordered_basis = [x for _, x in sorted(zip(inv_confs, basis),
                                                  key=lambda x: x[0])]
        return ordered_basis

    def rank(self, spectrum: Spectrum, formula: str) -> Ranking:
        ranking = self.run_metric(spectrum, formula)
        set_orig(ranking)
        val = 2**64
        basis = self.flitsr(spectrum, formula)
        spectrum.reset()
        if (not basis == []):
            ordered_basis = self.flitsr_ordering(spectrum, basis, ranking,
                                                 self.order_method)
            for x in ranking:
                if (x.group in basis):
                    x.score = val - ordered_basis.index(x.group)
            val = val-len(basis)
        # Reset the coverage matrix and counts
        ranking.sort(True)
        unset_orig()
        return ranking


@print_name('flitsr_multi')
class Multi(Flitsr):
    """
    Run the FLITSR* algorithm over the spectrum to produce ranked lists.
    """
    @existing('args')
    def __init__(self, args: 'Args'):
        flitsr_opts = args.get_arg_group('FLITSR')
        super().__init__(**flitsr_opts)

    def multiRemove(self, spectrum: Spectrum,
                    faulty: List[Spectrum.Group]) -> bool:
        """
        Remove the elements given by faulty from the spectrum, and remove any
        test cases executing these elements only.
        """
        # Get tests executing elems in faulty set
        executing: Set[Spectrum.Test] = set()
        for elem in faulty:
            exe = spectrum.get_tests(elem, only_failing=True)
            executing.update(exe)

        # Remove all elements in faulty set
        for group in faulty:
            spectrum.remove_group(group)

        multiFault = False
        for test in executing:
            for group in spectrum.groups():  # remaining groups not in faulty
                if (spectrum[test][group]):
                    break
            else:
                multiFault = True
                spectrum.remove(test, hard=True)
        return multiFault

    def rank(self, spectrum: Spectrum, formula: str) -> Ranking:
        ranking = self.run_metric(spectrum, formula)
        set_orig(ranking)
        val = 2**64
        newSpectrum = copy.deepcopy(spectrum)
        while (newSpectrum.tf > 0):
            basis = self.flitsr(newSpectrum, formula)
            if (not basis == []):
                ordered_basis = self.flitsr_ordering(spectrum, basis, ranking,
                                                     self.order_method)
                for x in ranking:
                    if (x.group in basis):
                        x.score = val - ordered_basis.index(x.group)
                val = val-len(basis)
            else:  # (fall-back) finish FLITSR* if basis is empty
                break
            # Reset the coverage matrix and counts
            newSpectrum.reset()
            # Next iteration can be either multi-fault, or multi-explanation
            # multi-fault -> we assume multiple faults exist
            # multi-explanation -> we assume there are multiple explanations
            # for the same faults
            self.multiRemove(newSpectrum, basis)
            val = val-1
        ranking.sort(True)
        unset_orig()
        return ranking



from flitsr.suspicious import Suspicious
from flitsr.ranking import Ranking, Rank
from flitsr.spectrum import Spectrum
from typing import List, Dict, Any


def basis(basis_num: int, spectrum: Spectrum,
          faults: Dict[int, List[Spectrum.Element]], ranking: Ranking,
          formula: str, effort: str) -> Ranking:
    new_ranking: Ranking = Ranking()
    r_iter = iter(ranking)
    rank = next(r_iter, None)
    first_fault = -1
    seen_basis = 0
    while (rank is not None and (seen_basis < basis_num or
                                 first_fault == -1)):
        temp_items: List[Rank] = []
        next_rank = rank
        while (next_rank is not None and next_rank.score == rank.score):
            if (first_fault == -1 and any(next_rank.elem in fault_locs for
                                          fault_locs in faults.values())):
                first_fault = len(temp_items)
            temp_items.append(next_rank)
            next_rank = next(r_iter, None)
        if (effort == 'worst' or first_fault == -1):
            new_ranking.extend(temp_items)
        elif (effort == 'best'):
            new_ranking.extend(temp_items[:first_fault+1])
        elif (effort == 'resolve'):
            new_ranking.extend([temp_items[first_fault]])
        if (next_rank is not None and rank.score-1 != next_rank.score):
            seen_basis += 1
        rank = next_rank
    return new_ranking


def oba(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
        ranking: Ranking, formula: str, effort: str):
    return method(float('inf'), spectrum, faults, ranking, effort)


def mba_dominator(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
                  ranking: Ranking, formula: str, effort: str):
    sus = Suspicious(spectrum.tf, spectrum.tf, spectrum.tp, spectrum.tp)
    score = sus.execute(formula)
    return method(score, spectrum, faults, ranking, effort)


def mba_zombie(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
               ranking: Ranking, formula: str, effort: str):
    sus = Suspicious(0, spectrum.tf, 0, spectrum.tp)
    score = sus.execute(formula)
    return method(score, spectrum, faults, ranking, effort)


def mba_5_perc(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
               ranking: Ranking, formula: str, effort: str):
    size = 0
    for group in spectrum.groups():
        size += len(group.get_elements())
    return method(int(size*0.05), spectrum, faults, ranking, effort, True)


def mba_10_perc(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
                ranking: Ranking, formula: str, effort: str):
    size = 0
    for group in spectrum.groups():
        size += len(group.get_elements())
    return method(int(size*0.1), spectrum, faults, ranking, effort, True)


def mba_const_add(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
                  ranking: Ranking, formula: str, effort: str):
    tot_size = 0
    for group in spectrum.groups():
        tot_size += len(group.get_elements())
    sus = Suspicious(0, spectrum.tf, 0, spectrum.tp)
    zero = sus.execute(formula)
    new_ranking = Ranking()
    r_iter = iter(ranking)
    rank = next(r_iter, None)
    stop_i = float('inf')
    f_num = 0
    size = 0
    while (rank is not None and size+1 <= stop_i and (rank.score > zero or
                                                      f_num == 0)):
        next_rank = rank
        fault_num = 0
        while (next_rank is not None and (next_rank.score == rank.score)):
            if (any(next_rank.elem in fault_locs for fault_locs in
                    faults.values())):
                fault_num += 1
            new_ranking.extend([next_rank])
            size += len(next_rank.group.get_elements())
            next_rank = next(r_iter, None)
        if (fault_num != 0):  # should've stopped already: size <= stop_i
            # recalculate stop amount
            f_num += fault_num
            stop_i = size + tot_size*0.01
        rank = next_rank
    return new_ranking


def mba_optimal(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
                ranking: Ranking, formula: str, effort: str):
    sus = Suspicious(0, spectrum.tf, 0, spectrum.tp)
    zero = sus.execute(formula)
    new_ranking = Ranking()
    r_iter = iter(ranking)
    rank = next(r_iter, None)
    stop_i = float('inf')
    f_num = 0
    size = 0
    while (rank is not None and size+1 <= stop_i and (rank.score > zero or
                                                      f_num == 0)):
        next_rank = rank
        fault_num = 0
        while (next_rank is not None and (next_rank.score == rank.score)):
            if (any(next_rank.elem in fault_locs for fault_locs in
                    faults.values())):
                fault_num += 1
            new_ranking.extend([next_rank])
            size += len(next_rank.group.get_elements())
            next_rank = next(r_iter, None)
        if (fault_num != 0):  # should've stopped already: size <= stop_i
            # recalculate stop amount
            f_num += fault_num
            stop_i = size + size/(f_num+1)
        rank = next_rank
    return new_ranking


def aba(spectrum: Spectrum, faults: Dict[int, List[Spectrum.Element]],
        ranking: Ranking, formula: str, effort: str):
    new_ranking = Ranking()
    r_iter = iter(ranking)
    rank = next(r_iter, None)
    temp_ranking: List[Rank] = []
    while (rank is not None and rank.score > 0.0):
        fault = False
        next_rank = rank
        while (next_rank is not None and next_rank.score == rank.score):
            if (any(next_rank.elem in fault_locs for fault_locs in
                    faults.values())):
                fault = True
            temp_ranking.append(next_rank)
            next_rank = next(r_iter, None)
        if (fault):
            new_ranking.extend(temp_ranking)
            temp_ranking.clear()
        rank = next_rank
    return new_ranking


def method(stop_score: float, spectrum: Spectrum,
           faults: Dict[int, List[Spectrum.Element]], ranking: Ranking,
           effort: str, by_rank=False):
    new_ranking = Ranking()
    r_iter = iter(ranking)
    rank = next(r_iter, None)
    first_fault = -1
    size = 0
    while (rank is not None and
           ((not by_rank and rank.score > stop_score)
            or (by_rank and size < stop_score) or first_fault == -1)):
        temp_ranking: List[Rank] = []
        next_rank = rank
        while (rank is not None and next_rank.score == rank.score):
            if (first_fault == -1 and any(next_rank.elem in fault_locs for
                                          fault_locs in faults.values())):
                first_fault = len(temp_ranking)
            temp_ranking.append(next_rank)
            size += len(next_rank.group.get_elements())
            next_rank = next(r_iter, None)
        if (effort == 'worst' or rank.score > stop_score or first_fault == -1):
            new_ranking.extend(temp_ranking)
        elif (effort == 2):
            new_ranking.extend(temp_ranking[:first_fault+1])
        else:
            new_ranking.extend([temp_ranking[first_fault]])
        rank = next_rank
    return new_ranking


# Populate the function names
funcs: Dict[str, Any] = {}
funcs.update([(name, func) for (name, func) in locals().items() if
              callable(func) and func.__module__ == __name__])


def getNames():
    all_names = funcs.keys()
    names = [x for x in all_names if (not x.startswith("__")
             and x != "cut" and x != "getNames" and x != "method")]
    return names


def cut(cutoff, spectrum, faults: Dict[int, List[Spectrum.Element]],
        ranking: Ranking, formula: str, effort: str):
    # get the function
    func = funcs[cutoff]
    return func(spectrum, faults, ranking, formula, effort)



from typing import List, Dict, Any, Set, Tuple
from flitsr.spectrum import Spectrum, Outcome


class SpectrumBuilder:
    def __init__(self):
        self._elements: List[Spectrum.Element] = []
        self._tests: List[Spectrum.Test] = []
        self._executions: Dict[Spectrum.Test, Set[Spectrum.Element]] = {}

    def get_tests(self):
        return self._tests

    def addTest(self, name: str, index: int, outcome: Outcome):
        """
        Add a new test to the spectrum with the given name, index, and outcome.

        :param name: the name of the test
        :param index: a unique identifier for this test (this may be modified)
        :param outcome: the outcome (pass/fail) of the test
        """
        t = Spectrum.Test(name, index, outcome)
        self._tests.append(t)
        self._executions[t] = set()
        return t

    def addElement(self, details: List[str],
                   faults: List[Any], index: int = None) -> Spectrum.Element:
        """
        Add a new element to the spectrum, with the given details and faults.
        """
        if (index is None):
            index = len(self._elements)
        e = Spectrum.Element(details, index, faults)
        self._elements.append(e)
        return e

    def addExecution(self, test: Spectrum.Test, elem: Spectrum.Element):
        """
        Mark the specified element as executed in the given test.
        """
        if (test not in self._executions):
            self._executions[test] = set()
        self._executions[test].add(elem)

    def addNonExecution(self, test: Spectrum.Test, elem: Spectrum.Element):
        """
        Mark the specified element as not executed in the given test.
        NOTE: Adding non-executions is not required for the spectrum
        (non-execution is assumed by default), thus this method does not do
        anything.
        """
        # if (test not in self._executions):
        #     self._executions[test] = {}
        # self._executions[test][elem] = False
        pass

    def form_groups(self) -> List[Spectrum.Group]:
        """
        Given one test pertaining to a row in the spectrum, split the groups
        according to the coverage. The split will potentially split each group
        in half, based on those elements that are executed in the test, and
        those that are not.
        """
        stack: List[Tuple[Set[Spectrum.Element], int]] = []
        groups: List[Spectrum.Group] = []
        # Initialize stack with all elements in one group
        stack.append((set(self._elements), 0))
        while (len(stack) > 0):
            cur, t_ind = stack.pop()
            if (t_ind >= len(self._tests) or len(cur) == 1):
                groups.append(Spectrum.Group(list(cur)))
            else:
                test = self._tests[t_ind]
                exe_elem = cur.intersection(self._executions[test])
                nexe_elem = cur.difference(exe_elem)
                if (len(exe_elem) != 0):
                    stack.append((exe_elem, t_ind+1))
                if (len(nexe_elem) != 0):
                    stack.append((nexe_elem, t_ind+1))
        return groups

    def get_spectrum(self) -> Spectrum:
        """ Return the computed spectrum """
        groups = self.form_groups()
        spectrum = Spectrum(self._elements, groups, self._tests,
                            self._executions)
        return spectrum


class SpectrumUpdater(SpectrumBuilder):
    def __init__(self, spectrum: Spectrum):
        super().__init__()
        self._elements = spectrum._elements[:]
        self._tests = spectrum.tests()[:]
        self._executions = dict()
        # copy all the executions
        for test in self._tests:
            new_exe = self._executions.setdefault(test, set())
            for group in spectrum.groups():
                # Only copy over actually executed elements
                if (spectrum[test][group]):
                    for elem in group.get_elements():
                        new_exe.add(elem)

    def remove_test_with_executions(self, test: Spectrum.Test):
        self._tests.remove(test)
        del self._executions[test]

    def copy_test_and_execution(self, test: Spectrum.Test, spectrum: Spectrum):
        self._tests.append(test)
        new_exe = self._executions.setdefault(test, set())
        for group in spectrum.groups():
            # Only copy over actually executed elements
            if (spectrum[test][group]):
                for other in group.get_elements():
                    # get this spectrum's element
                    elem = self._get_element(other)
                    new_exe.add(elem)

    def _get_element(self, other: Spectrum.Element):
        for elem in self._elements:
            if (elem.semantic_eq(other)):
                return elem
        raise KeyError(f'Could not find element {other} in SpectrumUpdater')



"""
This is a custom driver for running flitsr that facilitates the flitsr basis
ordering experiments.

The driver runs flitsr with each of its basis ordering types and stores each
result. The driver mostly mimicks the main method in the flitsr module (i.e.
the main entry point of flitsr), and uses the same command line arguments. The
main difference is the restriction to only running techniques with flitsr
activated (no base SBFL techniques) and running all variants of the flitsr
basis orderings on these.
"""


import re
import sys
from os import path as osp
from typing import List
from flitsr.args import Args
from flitsr.main import output, compute_cutoff
from flitsr.ranking import Ranking
from flitsr.input.input_reader import Input
from flitsr.errors import error
from flitsr.advanced import Config, RankerType, ClusterType


def main(argv: List[str]):
    args: Args = Args().parse_args(argv)
    # If only a ranking is given, print out metrics and return
    if (args.ranking):
        from flitsr.read_ranking import read_any_ranking
        ranking, spectrum = read_any_ranking(args.input,
                                             method_level=args.method)
        output([ranking], spectrum, args.weff, args.top1, args.perc_at_n,
               args.prec_rec, args.faults, args.collapse, csv=args.csv,
               specCsv=args.spectrum_csv,
               decimals=args.decimals, file=args.output)
        return
    # Else, run the full process
    try:
        reader = Input.get_reader(args.input)
    except ValueError as e:
        error(e)
    d_p = reader.get_run_file_name(args.input)
    # Read the spectrum in and setup parallel if needed
    spectrum = reader.read_spectrum(args.input, args.split, args.method)
    if (spectrum is None or len(spectrum.spectrum) == 0):
        print("ERROR: Incorrectly formatted input file, terminating...",
              file=sys.stderr)
        return
    # Execute techniques
    for config in [Config(RankerType['FLITSR']),
                   Config(RankerType['MULTI'])]:
        if (config.ranker is RankerType['FLITSR'] or
            config.ranker is RankerType['MULTI']):
            orderings = ['auto', 'conf', 'original', 'reverse', 'flitsr']
        else:
            orderings = ['original']
        for ordering in orderings:
            for metric in args.metrics:
                # Get the output channel
                input_filename = osp.basename(d_p)
                filename = (config.get_file_name() + '_' + ordering
                            + '_' + metric + '_' + input_filename)
                try:
                    output_file = open(filename, "x")
                except FileExistsError:
                    if (args.no_override):
                        # print("WARNING: Skipping execution of already "
                        #       "existing file", filename, file=sys.stderr)
                        continue
                    else:
                        print("WARNING: overriding file", filename,
                              file=sys.stderr)
                        output_file = open(filename, 'w')
                # Check for clustering
                if (config.cluster is not None or
                    hasattr(ClusterType, metric.upper())):
                    if (config.cluster is None):
                        cluster = ClusterType[metric.upper()]
                        # Set default metric for clustering
                        metric = 'ochiai'
                    else:
                        cluster = config.cluster
                    cluster_params = args.get_arg_group(cluster.name)
                    cluster_mthd = cluster.value(**cluster_params)
                    spectrums = cluster_mthd.cluster(args.input, spectrum,
                                                     args.method)
                else:
                    spectrums = [spectrum]
                rankings: List[Ranking] = []
                # Run each sub-spectrum
                for subspectrum in spectrums:
                    # Run techniques
                    ranker = config.ranker
                    if (ranker is None):
                        ranker = RankerType['SBFL']
                    if (ranker == RankerType['SBFL'] and
                        hasattr(RankerType, metric.upper())):
                        ranker = RankerType[metric.upper()]
                        metric = 'ochiai'
                    ranker_params = args.get_arg_group(ranker.name)
                    if ('internal_ranking' in ranker_params):
                        ranker_params['internal_ranking'] = ordering
                    ranker_mthd = ranker.value(**ranker_params)
                    ranking = ranker_mthd.rank(subspectrum, metric)
                    # Compute cut-off
                    if (args.cutoff_strategy):
                        ranking = compute_cutoff(args.cutoff_strategy, ranking,
                                                 subspectrum, metric,
                                                 args.cutoff_eval)
                    rankings.append(ranking)
                # Compute and print output
                output(rankings, spectrum, args.weff, args.top1, args.perc_at_n,
                       args.prec_rec, args.faults, args.collapse, csv=args.csv,
                       specCsv=args.spectrum_csv, decimals=args.decimals, file=output_file)
                spectrum.reset()


if __name__ == "__main__":
    main(sys.argv[1:])



from matplotlib import pyplot as plt # type: ignore
from matplotlib import ticker as mtick
import re
import os
import sys

if __name__ == "__main__":
    seperate = False
    rel = False
    tcm = True
    custom = None
    metrics = []
    i = 1
    while (True):
        if (len(sys.argv) > i):
            if (sys.argv[i] == "sep"):
                seperate = True
            elif (sys.argv[i] == "rel"):
                rel = True
            elif (sys.argv[i] == "tcm"):
                tcm = True
            elif (sys.argv[i].startswith("calcs=[")):
                index = sys.argv[i].index("[")+1
                custom = [x.strip() for x in sys.argv[i][index:-1].split(",")]
            elif (sys.argv[i].startswith("metrics=[")):
                index = sys.argv[i].index("[")+1
                metrics = [x.strip() for x in sys.argv[i][index:-1].split(",")]
            else:
                print("Unknown option:", sys.argv[i])
                exit(1)
            i += 1
        else:
            break
    faults = []
    total = 0
    num_locs = 0
    if (os.path.isfile("size")):
        num_locs = int(open("size").readline())
    proj = os.getcwd().split("/")[-1].capitalize()
    for d in os.scandir():
        if (tcm and d.is_dir() and d.name.endswith("_Chart")):
            num = int(d.name.split('_')[0])
            faults.append(num)
        elif (d.is_dir() and d.name.endswith("-fault")):
            num = int(d.name.split('-')[0])
            faults.append(num)
    faults.sort()
    names = {}
    size = 0
    for f in faults:
        if (rel):
            if (tcm):
                file = open(str(f)+"-fault/results_rel")
            else:
                file = open(str(f)+"_fault_versions_Chart/results_rel")
        else:
            if (tcm):
                file = open(str(f)+"-fault/results")
            else:
                file = open(str(f)+"_fault_versions_Chart/results")
        line = file.readline()
        while (not line == ""):
            metric = line.strip()
            if (not metric in names):
                names[metric] = {}
            name = names[metric]
            line = file.readline()
            while (line.count("\t") == 1):
                mode = line.strip()
                if (not mode in name):
                    name[mode] = {}
                mod = name[mode]
                line = file.readline()
                while (line.count("\t") == 2):
                    calc = line.strip().split(": ")
                    if (not calc[0] in mod):
                        mod[calc[0]] = []
                        size += 1
                    if (rel):
                        mod[calc[0]].append(float(calc[1][:-1]))
                    else:
                        mod[calc[0]].append(float(calc[1]))
                    line = file.readline()
    if (len(metrics) == 0):
        metrics = names.keys()
    color = ['r', 'b', 'g', 'y', 'm', 'c', 'k']
    shape = ['^', '.', 'v', '-']
    if (seperate):
        if (custom):
            size = len(custom)
        fig, axs = plt.subplots(size,len(metrics), sharey='row',
                gridspec_kw={"left": 0.022,
                 "bottom": 0.027,
                 "right":0.994,
                 "top": 0.932,
                 "wspace": 0.09,
                 "hspace": 0.131})
        row = size
    else:
        fig, axs = plt.subplots(2,2, sharey=True, sharex=True,
                gridspec_kw={"left": 0.035,
                 "bottom": 0.047,
                 "right":0.98,
                 "top": 0.90,
                 "wspace": 0.126,
                 "hspace": 0.215})
        row = 2
    ax = 0

    #remove unused things
    #del names['jaccard']
    for metric in metrics:
        #print(metric)
        modes = names[metric]
        i = 0
        for mode in modes.keys():
            #print(mode)
            calcs = modes[mode]
            if (not custom):
                custom = calcs.keys()
            j = 0
            for calc in custom:
                values = calcs[calc]
                #print(faults, values)
                if (seperate):
                    axs[j, ax].plot(faults, values, str(color[i])+".-")
                    #if (rel):
                        #axs[j, ax].set_title("Comparison of the relative wasted"
                                #+" effort for the \n"+calc +" fault using the "+
                                #metric+" metric")
                    #else:
                        #axs[j, ax].set_title("Comparison of the wasted effort for the "+calc
                            #+"\nfault using the "+metric+" metric")
                    axs[j, ax].set_title(metric+" "+calc)
                    axs[j, ax].grid(True)
                    #legendloc = "upper left" if j == 1 else "upper right"
                    axs[j, ax].legend([x for x in list(modes.keys())],
                            prop={"size":12})
                else:
                    axs[int(ax/row), ax%row].plot(faults, values, str(color[i])+".-")
                j += 1
            i += 1
        if (not seperate):
            if (rel):
                axs[int(ax/row), ax%row].set_title("Comparison of the relative wasted"+
                    "effort for the first, average and med faults using the "+
                    metric+"\nmetric")
            else:
                axs[int(ax/row), ax%row].set_title("Comparison of the wasted"+
                    "effort for the first, average and med faults using the "+
                    metric+"\nmetric")
            axs[int(ax/row), ax%row].grid()
            axs[int(ax/row), ax%row].legend([x+" "+y for x in list(modes.keys())
                    for y in list(calcs.keys())], prop={"size":6})
        ax += 1

    if (rel):
        fmt = '%.0f%%' # Format you want the ticks, e.g. '40%'
        yticks = mtick.FormatStrFormatter(fmt)
        for a in axs:
            for ax in a:
                ax.yaxis.set_major_formatter(yticks)


    fig.suptitle(proj+" (size: "+str(num_locs)+")", fontsize=18)
    plt.show()



import sys
import re
import os
from os import path as osp
from io import TextIOWrapper
from shutil import rmtree
from typing import Dict, Tuple, List
from flitsr.split_faults import split
from flitsr.spectrum import Spectrum, Outcome
from flitsr.spectrumBuilder import SpectrumBuilder
from flitsr import errors
from flitsr.input.input_reader import Input


class Gzoltar(Input):
    def get_run_file_name(self, input_path: str):
        return input_path.split("/")[0] + ".run"

    def construct_details(self, f: TextIOWrapper, method_level: bool,
                          sb: SpectrumBuilder) -> Dict[int, Spectrum.Element]:
        """
        Fills the spectrum object with elements read in from the open file 'f'.
        """
        i = 0  # number of actual lines
        method_map: Dict[int, Spectrum.Element] = {}
        methods: Dict[Tuple[str, str], Spectrum.Element] = {}
        bugs = 0
        assert f.readline() == 'name\n'  # remove header
        for line in f:
            m = re.fullmatch('([^$]+)\\$([^#]+)#([^:]+):([0-9]+)(?::(.+))?',
                             line.rstrip())
            if (m is None):
                errors.error("Incorrectly formatted line \"" + line +
                             "\" when reading input file")
            else:
                faults = []
                if (m.group(5) is not None):
                    bs = m.group(5).split(':')
                    if (not bs[0].isdigit()):
                        faults = [bugs]
                    else:
                        faults = []
                        for b in bs:
                            faults.append(int(b))
                    bugs += 1
                if (method_level):
                    details = [m.group(1)+"."+m.group(2), m.group(3), m.group(4)]
                    if ((details[0], details[1]) not in methods):
                        # add with first line number
                        elem = sb.addElement(details, faults)
                        methods[(details[0], details[1])] = elem
                        method_map[i] = elem
                    else:
                        elem = method_map[i] = methods[(details[0], details[1])]
                        # update method's faults
                        for fault in faults:
                            if (fault not in elem.faults):
                                elem.faults.append(fault)
                else:
                    details = [m.group(1)+"."+m.group(2), m.group(3), m.group(4)]
                    elem = sb.addElement(details, faults)
                    method_map[i] = elem
                i += 1
        return method_map

    def construct_tests(self, tests_reader: TextIOWrapper, sb: SpectrumBuilder):
        tests_reader.readline()
        i = 0
        for line in tests_reader:
            m = re.fullmatch('([^,]+),(PASS|FAIL|ERROR)(,.*)?', line.rstrip())
            if (m is None):
                errors.error("incorrectly formatted test line in input file:",
                             line, "terminating...")
            else:
                sb.addTest(m.group(1), i, Outcome[m.group(2)])
            i += 1

    def fill_spectrum(self, bin_file: TextIOWrapper,
                      method_map: Dict[int, Spectrum.Element],
                      sb: SpectrumBuilder):
        for t, test in enumerate(sb.get_tests()):
            line = bin_file.readline()
            if (line == ''):
                errors.error('Incorrect number of matrix lines', f'({t})',
                             'in input file, terminating...')
            arr = line.rstrip().split()
            seen = set()
            # Loop over all elements in test exec (remove test outcome at the end)
            for i in range(0, len(arr)-1):
                try:
                    elem = method_map[i]
                except KeyError:
                    errors.error('Incorrect number of matrix columns', f'({i})',
                                 'in input file, terminating...')
                if (arr[i] != "0" and elem not in seen):
                    sb.addExecution(test, elem)
                    seen.add(elem)

    def read_spectrum(self, input_path: str, split_faults: bool,
                      method_level=False) -> Spectrum:
        sb = SpectrumBuilder()
        # Getting the details of the elements
        method_map = self.construct_details(open(input_path+"/spectra.csv"),
                                            method_level, sb)
        # Getting the details of the tests
        self.construct_tests(open(input_path+"/tests.csv"), sb)
        # Constructing the spectrum
        self.fill_spectrum(open(input_path+"/matrix.txt"), method_map, sb)
        spectrum = sb.get_spectrum()
        del sb
        # Split fault groups if necessary
        if (split_faults):
            faults, unexposed = split(spectrum.get_faults(), spectrum)
            for elem in unexposed:
                elem.faults.clear()
                print("WARNING: Dropped faulty UUT:", elem, "due to unexposure",
                      file=sys.stderr)
            # Get element's fault lists
            fault_lists: Dict[Spectrum.Element, List[float]] = {}
            for (f_num, f_locs) in faults.items():
                for elem in f_locs:
                    fault_lists.setdefault(elem, []).append(f_num)
            # Set element's fault lists
            for (elem, f_list) in fault_lists.items():
                elem.faults = f_list
            if (len(faults) == 0):
                print("ERROR: No exposable faults in ", input_path, ", terminating...",
                        sep="", file=sys.stderr)
                quit()
        return spectrum

    @staticmethod
    def check_format(input_path: str) -> bool:
        return (osp.isdir(input_path) and
                osp.isfile(osp.join(input_path, "matrix.txt")) and
                osp.isfile(osp.join(input_path, "spectra.csv")) and
                osp.isfile(osp.join(input_path, "tests.csv")))

    @classmethod
    def write_spectrum(cls, spectrum: Spectrum, directory: str):
        """ Output the spectrum in Gzoltar format """
        type_ = cls.get_type()
        # First check that the directory exists and is empty
        if (not osp.isdir(directory)):
            os.mkdir(directory)
        else:
            rmtree(directory)
            os.mkdir(directory)
        # Next print out the tests
        with open(osp.join(directory, "tests.csv"), 'w') as test_file:
            print("name,outcome,runtime,stacktrace", file=test_file)
            for test in spectrum.tests():
                print(test.name, "PASS" if test.outcome is Outcome.PASSED else
                      "FAIL", "", "", sep=",", file=test_file)
        with open(osp.join(directory, "spectra.csv"), 'w') as units_file:
            print("name", file=units_file)
            # TODO: change _elements below to elements()
            for elem in spectrum._elements:
                print(elem.output_str(type_=type_), file=units_file)
        with open(osp.join(directory, "matrix.txt"), 'w') as matrix_file:
            for test in spectrum.tests():
                for elem in spectrum._elements:
                    print(int(spectrum[test][elem]), end=" ", file=matrix_file)
                print('+' if test.outcome is Outcome.PASS else '-',
                      file=matrix_file)

    @classmethod
    def get_elem_separators(cls):
        return ['$', '#', ':', ':']


if __name__ == "__main__":
    from flitsr.output import print_spectrum
    d = sys.argv[1]
    ginput = Gzoltar()
    spectrum = ginput.read_spectrum(d, False)
    print_spectrum(spectrum)



# PYTHON_ARGCOMPLETE_OK
import sys
from os import path as osp
from typing import List, Optional
from argparse import ArgumentParser
from enum import Enum
from flitsr.errors import error
from flitsr.input.input_reader import Input
from flitsr.input import InputType


def convert(input_file: str, output_format: Optional[InputType] = None,
            output_file: Optional[str] = None):
    """Convert a input file from one spectral format to another"""
    # Check the input file type and set input method
    try:
        reader = Input.get_reader(input_file)
    except ValueError as e:
        error(e)
    input_format = reader.get_type()
    spectrum = reader.read_spectrum(input_file, False)
    # Set the output method if necessary
    if (output_format is None):
        if (input_format is InputType['TCM']):
            output_format = InputType['GZOLTAR']
        else:
            output_format = InputType['TCM']
    # Set the output file name if necessary
    if (output_file is None):
        output_file = osp.splitext(input_file)[0] + (".tcm" if output_format is
                                                     InputType['TCM'] else "")
    if (output_file == input_file):
        error(f"Cannot override input file {input_file} with output")
    # Print out the spectrum in the converted format
    output_format.value.write_spectrum(spectrum, output_file)


def main(argv: Optional[List[str]] = None):
    parser = ArgumentParser('transform', description='Transform a spectrum '
                            'from one format to another')
    parser.add_argument('input', help='The spectrum input file. The spectral '
                        'format is guessed from the input file, and by '
                        'default the output format is assumed to be a '
                        'different format type.')
    output_types = [it.name for it in list(InputType)]
    parser.add_argument('-t', '--output-type', help='Convert the input file'
                        'into the given output format', choices=output_types)
    parser.add_argument('-o', '--output-file', help='Specifies the name of '
                        'the output file. By default, the name is derived '
                        'from the name of the input file; for example, if the '
                        'input file is "input.tcm", and the output format is '
                        'Gzoltar, the output will be "input/".')
    args = parser.parse_args(argv)
    if (args.output_type is not None):
        args.output_type = InputType[args.output_type]
    convert(args.input, args.output_type, args.output_file)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from flitsr.spectrum import Spectrum
from flitsr import advanced


class Refiner(ABC):
    @abstractmethod
    def refine(self, spectrum: Spectrum, method_lvl=False) -> Spectrum:
        pass

    def __init_subclass__(cls):
        advanced.register_refiner(cls)



from abc import ABC, abstractmethod
from typing import List
from flitsr.spectrum import Spectrum
from flitsr import advanced


class Cluster(ABC):
    @abstractmethod
    def cluster(self, inp_file: str, spectrum: Spectrum,
                method_lvl=False) -> List[Spectrum]:
        pass

    def __init_subclass__(cls):
        advanced.register_cluster(cls)



from flitsr.tie import Ties, Tie


def one_top1(ties: Ties) -> bool:
    tie = get_top1(ties)
    for fault in ties.faults.values():
        for loc in fault:
            if (loc in tie.elems()):
                return True
    return False


def all_top1(ties: Ties) -> bool:
    tie = get_top1(ties)
    count = 0
    for fault in ties.faults.values():
        for loc in fault:
            if (loc in tie.elems()):
                count += 1
                break  # Only consider first location of fault
    return (count == len(ties.faults))


def percent_top1(ties: Ties) -> float:
    if (len(ties.faults) == 0):
        return 100
    else:
        tie = get_top1(ties)
        count = 0
        for fault in ties.faults.values():
            for loc in fault:
                if (loc in tie.elems()):
                    count += 1
                    break
        return (count/len(ties.faults))*100


def size_top1(ties: Ties) -> int:
    tie = get_top1(ties)
    return len(tie.elems())


# <------------------------- Helper functions ---------------------->


def get_top1(ties: Ties) -> Tie:
    tie_iter = iter(ties)
    return next(tie_iter)



from flitsr.ranking import Ranking, Tiebrk
from flitsr.advanced.ranker import Ranker
from flitsr.advanced.attributes import existing, print_name
from flitsr.spectrum import Spectrum
from flitsr.suspicious import Suspicious


@print_name('base')
class SBFL(Ranker):
    """
    Disables the FLITSR algorithm so that only the base metric
    is used to produce the ranking. This is equivalent to using the
    base metric as-is, but allows the user to run these metrics
    within the FLITSR framework
    """

    @existing('tiebrk')
    def __init__(self, tiebrk: Tiebrk = Tiebrk.ORIG):
        self.tiebrk = tiebrk

    def rank(self, spectrum: Spectrum, base_metric: str) -> Ranking:
        return Suspicious.apply_formula(spectrum, base_metric,
                                        tiebrk=self.tiebrk)



# PYTHON_ARGCOMPLETE_OK
import sys
import re
import copy
from os import path as osp
from math import log
from typing import List, Optional
from flitsr import weffort
from flitsr import top
from flitsr import percent_at_n
from flitsr import precision_recall
from flitsr.output import print_csv, print_spectrum_csv, print_names
from flitsr import cutoff_points
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking
from flitsr.tie import Ties
from flitsr.args import Args
from flitsr.advanced import ClusterType, RankerType
from flitsr.input.input_reader import Input
from flitsr.errors import error


def compute_cutoff(cutoff: str, ranking: Ranking, spectrum: Spectrum,
                   mode: str, effort: str) -> Ranking:
    faults = spectrum.get_faults()
    if (cutoff.startswith("basis")):
        if ('=' in cutoff):
            num_bases = int(cutoff.split('=')[1])
        else:
            num_bases = 1
        ranking = cutoff_points.basis(num_bases, spectrum, faults, ranking,
                                      mode, effort=effort)
    else:
        ranking = cutoff_points.cut(cutoff, spectrum, faults, ranking, mode,
                                    effort=effort)
    return ranking


def output(rankings: List[Ranking], spectrum: Spectrum, weff=[], top1=[],
           perc_at_n=[], prec_rec=[], faults=[], collapse=False,
           csv=False, specCsv=False, decimals=2, file=sys.stdout):
    if (weff or top1 or perc_at_n or prec_rec or faults):
        ties: Ties = Ties(spectrum, rankings)
        if (weff):
            if ("first" in weff):
                print("wasted effort (first): {:.{}f}".format(
                    weffort.first(ties, collapse),
                    decimals), file=file)
            if ("avg" in weff):
                print("wasted effort (avg): {:.{}f}".format(
                    weffort.average(ties, collapse),
                    decimals), file=file)
            if ("med" in weff):
                print("wasted effort (median): {:.{}f}".format(
                    weffort.median(ties, collapse),
                    decimals), file=file)
            if ("last" in weff):
                print("wasted effort (last): {:.{}f}".format(
                    weffort.last(ties, collapse),
                    decimals), file=file)
            for nth in [w for w in weff if type(w) == int]:
                print("wasted effort ({}): {:.{}f}".format(
                    nth, weffort.nth(ties, int(nth), collapse),
                    decimals), file=file)
        if (top1):
            if ("one" in top1):
                print("at least 1 ranked #1: {:.{}f}".format(
                    top.one_top1(ties),
                    decimals), file=file)
            if ("all" in top1):
                print("all ranked #1: {:.{}f}".format(
                    top.all_top1(ties),
                    decimals), file=file)
            if ("perc" in top1):
                print("percentage ranked #1: {:.{}f}".format(
                    top.percent_top1(ties),
                    decimals), file=file)
            if ("size" in top1):
                print("size of #1: {:.{}f}".format(
                    top.size_top1(ties),
                    decimals), file=file)
        if (perc_at_n):
            bumps = percent_at_n.getBumps(ties, spectrum, collapse=collapse)
            if ('perc' in perc_at_n):
                form = ','.join(['{{:.{}f}}'.format(decimals)]*len(bumps))
                print("percentage at n:", form.format(*bumps), file=file)
            if (any([a in perc_at_n for a in ['auc', 'pauc', 'lauc']])):
                auc = percent_at_n.auc_calc(
                        percent_at_n.combine([(bumps[0], bumps[1:])]))
                if ('auc' in perc_at_n):
                    print("auc:", auc, file=file)
                if ('pauc' in perc_at_n):
                    optimal = percent_at_n.auc_calc([(0.0, 100.0)])
                    print("pauc:", "{:.{}f}".format(auc/optimal, decimals),
                          file=file)
                if ('lauc' in perc_at_n):
                    optimal = percent_at_n.auc_calc([(0.0, 100.0)])+1
                    lauc = abs(1-log(optimal-auc, optimal))
                    print("lauc:", "{:.{}f}".format(lauc, decimals), file=file)
        if (prec_rec):
            for entry in prec_rec:
                if (entry[0] == 'p'):
                    p = precision_recall.precision(entry[1], ties, spectrum,
                                                   collapse)
                    print("precision at {}: {:.{}f}".format(entry[1], p,
                                                            decimals), file=file)
                elif (entry[0] == 'r'):
                    r = precision_recall.recall(entry[1], ties, spectrum,
                                                collapse)
                    print("recall at {}: {:.{}f}".format(entry[1], r,
                                                         decimals), file=file)
        if (faults):
            if ("num" in faults):
                print("fault number: {}".format(len(ties.faults)), file=file)
            if ("ids" in faults):
                print("fault ids: {}".format(list(ties.faults.keys())),
                      file=file)
            if ("elems" in faults):
                print("fault elements: {}".format([e for es in ties.faults.values()
                                                   for e in es]), file=file)
            if ("all" in faults):
                print("fault info: {}".format(ties.faults), file=file)
    elif (csv):
        for (i, s) in enumerate(rankings):
            if (i > 0):
                print('<', '-'*22, ' Next Ranking ', '-'*22, '>', sep='',
                      file=file)
            print_csv(spectrum, s, file)
    elif (specCsv):
        print_spectrum_csv(spectrum, file)
    else:
        for (i, s) in enumerate(rankings):
            if (i > 0):
                print('<', '-'*22, ' Next Ranking ', '-'*22, '>', sep='',
                      file=file)
            print_names(spectrum, s, file)


def main(argv: Optional[List[str]] = None):
    args: Args = Args().parse_args(argv)
    # If only a ranking is given, print out metrics and return
    if (args.ranking):
        from flitsr.read_ranking import read_any_ranking
        ranking, spectrum = read_any_ranking(args.input,
                                             method_level=args.method)
        output([ranking], spectrum, args.weff, args.top1, args.perc_at_n,
               args.prec_rec, args.faults, args.collapse, csv=args.csv,
               specCsv=args.spectrum_csv,
               decimals=args.decimals, file=args.output)
        return
    # Else, run the full process
    try:
        reader = Input.get_reader(args.input)
    except ValueError as e:
        error(e)
    d_p = reader.get_run_file_name(args.input)
    # Read the spectrum in and setup parallel if needed
    gspectrum = reader.read_spectrum(args.input, args.split, args.method)
    if (gspectrum is None or len(gspectrum.spectrum) == 0):
        print("ERROR: Incorrectly formatted input file, terminating...",
              file=sys.stderr)
        return
    # Execute techniques
    for config in args.types:
        for metric in args.metrics:
            # Get the output channel
            if (len(args.metrics) == 1 and len(args.types) == 1 and not args.all):
                output_file = args.output
            else:
                # store output files in the current directory
                input_filename = osp.basename(d_p)
                filename = (config.get_file_name() + '_' + metric + '_'
                            + input_filename)
                try:
                    output_file = open(filename, "x")
                except FileExistsError:
                    if (args.no_override):
                        # print("WARNING: Skipping execution of already "
                        #       "existing file", filename, file=sys.stderr)
                        continue
                    else:
                        print("WARNING: overriding file", filename,
                              file=sys.stderr)
                        output_file = open(filename, 'w')
            # copy spectrum
            spectrum = copy.deepcopy(gspectrum)
            # Check for spectrum refining
            if (config.refiner is not None):
                refiner_params = args.get_arg_group(config.refiner.name)
                refiner_mthd = config.refiner.value(**refiner_params)
                spectrum = refiner_mthd.refine(spectrum, args.method)
            # Check for clustering
            if (config.cluster is not None or
                hasattr(ClusterType, metric.upper())):
                # deal with clustering technique as metric
                if (config.cluster is None):
                    cluster = ClusterType[metric.upper()]
                    # Set default metric for clustering
                    metric = 'ochiai'
                else:
                    cluster = config.cluster
                cluster_params = args.get_arg_group(cluster.name)
                cluster_mthd = cluster.value(**cluster_params)
                spectrums = cluster_mthd.cluster(args.input, spectrum,
                                                 args.method)
            else:
                spectrums = [spectrum]
            rankings: List[Ranking] = []
            # Run each sub-spectrum
            for subspectrum in spectrums:
                # Run techniques
                ranker = config.ranker
                if (ranker is None):
                    ranker = RankerType['SBFL']
                if (ranker == RankerType['SBFL'] and
                    hasattr(RankerType, metric.upper())):
                    ranker = RankerType[metric.upper()]
                    metric = 'ochiai'
                ranker_params = args.get_arg_group(ranker.name)
                ranker_mthd = ranker.value(**ranker_params)
                ranking = ranker_mthd.rank(subspectrum, metric)
                # Compute cut-off
                if (args.cutoff_strategy):
                    ranking = compute_cutoff(args.cutoff_strategy, ranking,
                                             subspectrum, metric,
                                             args.cutoff_eval)
                rankings.append(ranking)
            # Compute and print output
            output(rankings, spectrum, args.weff, args.top1, args.perc_at_n,
                   args.prec_rec, args.faults, args.collapse, csv=args.csv,
                   specCsv=args.spectrum_csv, decimals=args.decimals, file=output_file)
            spectrum.reset()


if __name__ == "__main__":
    main()



# PYTHON_ARGCOMPLETE_OK
import math
import sys
import ast
import numpy as np
from argparse import ArgumentParser
import argcomplete
from enum import Enum
from typing import Dict, List, Tuple, Optional
from flitsr.spectrum import Spectrum
from flitsr.tie import Ties
from flitsr.suspicious import Suspicious


def getBumps(ties: Ties, spectrum: Spectrum, worst_effort=False,
             collapse=False) -> List[float]:
    if (len(ties.faults) == 0):
        return [0.0]
    tie_iter = iter(ties)
    total = 0
    size = 0
    if (collapse):
        size = len(spectrum.groups())
    else:
        for group in spectrum.groups():
            size += len(group.get_elements())
    bumps = [float(size)]
    try:
        while (True):
            tie = next(tie_iter)
            for f in range(1, tie.num_faults()+1):
                expect_value = tie.expected_value(f, False, collapse)
                bumps.append(total+expect_value)
            total += tie.len(collapse)
    except StopIteration:
        pass
    return bumps


def combine(results: List[Tuple[float, List[float]]]) -> List[Tuple[float, float]]:
    def err(x):
        return max(10**-6, (math.log((99*x/100) + 1)/math.log(100))*0.1)
    size = len(results)
    total = 0.0
    pointers = [0]*size
    final = [len(r[1]) for r in results]
    combined = []
    def val(results, pointer, i):
        if (pointer < len(results[i][1])):
            return results[i][1][pointer]*100/results[i][0]
    while (pointers != final):
        # get the minimum next value
        min_ = min([v for v in (val(results, pointers[i], i) for i in range(size))
                        if v is not None])
        indexes = []
        for i in range(size):
            j = 0
            while ((v := val(results, pointers[i]+j, i)) is not None):
                if (abs(min_ - v) < err(min_)):
                    j += 1
                else:
                    break
            if (j > 0):
                indexes.append((i, j))
        #curr = [(math.inf if (pointers[i] == len(results[i])) else results[i][pointers[i]])
                #for i in range(size)]
        #min_ = min(curr)
        #indexes = [i for i,x in enumerate(curr) if x - min_ < 10e-4]
        for (i, j) in indexes:
            pointers[i] += j
            total += j * (100/(len(results[i][1])*size))
        combined.append((min_, round(total, 8)))
    return combined


plot_type = Enum("plot_type", "metric mode")


def read_comb_file(comb_file: str) -> Tuple[List[str], List[str],
        Dict[Tuple[str, str], List[Tuple[float, float]]]]:
    lines = open(comb_file).readlines()
    metrics = []
    modes = []
    metric = ""
    mode = ""
    points = {}
    for line in lines:
        if (line.startswith("\t\t")):
            comb = ast.literal_eval(line.strip().split(": ")[1])
            comb.append((100.0, 100.0))
            points[(metric, mode)] = comb
        elif (line.startswith("\t")):
            mode = line.strip()
            if (mode not in modes):
                modes.append(mode)
        else:
            metric = line.strip()
            if (metric not in metrics):
                metrics.append(metric)
    return metrics, modes, points


def plot(plot_file: str, log=True, all=False, type=plot_type.metric,
         metrics=None, flitsrs=None):
    from matplotlib import pyplot as plt
    modes: List[str] = []
    comb_points: Dict[Tuple[str, str], List[Tuple[float, float]]] = {}
    if (metrics is None):
        metrics, modes, comb_points = read_comb_file(plot_file)
    else:
        _, modes, comb_points = read_comb_file(plot_file)
    points = {}
    for item in comb_points.items():
        key = item[0]
        x = [0.0]
        y = [0.0]
        for inc in item[1]:
            x.append(inc[0])
            y.append(inc[1])
        points[key] = (x, y)
    if (all):
        num_plots = 1
    elif (type == plot_type.metric):
        num_plots = len(metrics)
    else:
        num_plots = len(modes)
    plt.rcParams.update({'font.size': 10})
    fig, axs = plt.subplots(1, num_plots,
                            gridspec_kw={"left": 0.045,
                                         "bottom": 0.06,
                                         "right": 0.99,
                                         "top": 0.99,
                                         "wspace": 0.2,
                                         "hspace": 0.2} if all else {})
    if (all):
        plot_all(axs, points, metrics, modes, flitsrs, log)
    else:
        plot_sep(axs, points, type, metrics, modes, log)
    plt.show()


def plot_all(axs, points, metrics, modes, flitsrs, log):
    import matplotlib.cm as cm
    color = list(cm.rainbow(np.linspace(0, 1, len(metrics))))
    style = [(), (6, 3), (1, 3), (1, 3, 6, 3), (3, 3), (6, 3, 3, 3)]
    marker = ['D', 'o', '^', '8', 's', 'p', '*', 'x', '+', 'v', '<', '>', 'P',
              'h', 'X', 'H', 'd', '|', '_']
    lines = []
    for (i, me) in enumerate(metrics):
        for (j, mo) in enumerate(modes):
            if ((me, mo) in points and (mo == "Base" or flitsrs is None
                                        or me in flitsrs)):
                point = points[(me, mo)]
                # print(point)
                ls = axs.plot(point[0], point[1], dashes=style[j],
                              color=color[i], marker=marker[i], markevery=0.1)
                if (mo == "Base"):
                    lines.append(ls[0])
    # lines = axs.get_lines()[::len(modes)]
    dummy_lines = []
    for j in range(len(modes)):
        dummy_lines.append(axs.plot([], [], c="black",
                                    dashes=style[j])[0])
    t_dummy, = axs.plot([], [], marker='None', ls='None')
    labels = [r'$\mathbf{Metrics\!:}$'] + metrics + \
             [r'$\mathbf{Types\!:}$'] + modes
    all_lines = [t_dummy] + lines + [t_dummy] + dummy_lines
    axs.legend(all_lines, labels)
    # axs.set_title("")
    plot_log(axs, log)
    # plt.ylim(0, 100)
    # plt.xlim(0, 100)
    axs.grid()


def plot_sep(axs, points, type, metrics, modes, log):
    if (type == plot_type.metric):
        split, merged = metrics, modes
    else:
        split, merged = modes, metrics
    for (i, s) in enumerate(split):
        for (j, m) in enumerate(merged):
            if (type == plot_type.mode and (m, s) in points):
                point = points[(m, s)]
            elif ((s, m) in points):
                point = points[(s, m)]
            # print(point)
            axs[i].plot(point[0], point[1])
    # plt.step(x,y)
    for i in range(len(axs)):
        ax = axs[i]
        ax.legend(merged)
        ax.set_title(split[i])
        plot_log(ax, log)
        # plt.ylim(0, 100)
        # plt.xlim(0, 100)
        ax.grid()


def plot_log(ax, log):
    if (log):
        ax.set_xscale("log")
        ax.set_xticks([0.01, 0.1, 1, 10, 100], [0.01, 0.1, 1, 10, 100])


def auc_calc(points: List[Tuple[float, float]],
             cut_off=101.0) -> float:
    points = sorted(points, key=lambda x: x[0])  # sort points, sanity check
    auc = 0.0
    if (len(points) == 0 or points[-1][0] != 100.0):
        points.append((100.0, 100.0))
    for i in range(1, len(points)):
        if (points[i][0] >= cut_off):
            break
        auc += (points[i][0] - points[i-1][0]) * points[i-1][1]
    return auc


def main(argv: Optional[List[str]] = None):
    parser = ArgumentParser(prog="percent_at_n")
    subparsers = parser.add_subparsers(title='Modes', description='The '
                                       'following modes are available '
                                       'for the percent_at_n command:',
                                       required=True, dest='mode')
    combine_parser = subparsers.add_parser('combine')
    combine_parser.add_argument('input_file')

    met_names = list(map(str.capitalize, Suspicious.getNames(True)))
    plot_parser = subparsers.add_parser('plot', help='Plots the percentage at '
                                        'n graph for the given input file '
                                        'produced by the merge script')
    plot_parser.add_argument('plot_file', help='The input file generated by '
                             'the FLITSR merge script')
    plot_parser.add_argument('-a', '--all', action='store_true',
                             help='Plots one graph containing all metrics and '
                             'advanced types. You may optionally filter the '
                             'metrics that are plotted and for which metrics '
                             'advanced types are plotted for by the --metrics '
                             'and --types options')
    plot_parser.add_argument('-m', '--metrics', nargs='+', choices=met_names,
                             help='Specifies the metrics to plot for the --all '
                             'plotting style. Allowed values are: '
                             f'[{", ".join(met_names)}]', metavar='METRIC')
    plot_parser.add_argument('-t', '--types', nargs='+', choices=met_names,
                             help='Specifies the metrics for which to plot '
                             'advanced types for using the --all plotting '
                             'style. See --metrics for allowed values',
                             metavar='METRIC')
    plot_parser.add_argument('-s', '--split-type', choices=['metric', 'type'],
                             help='Determines whether to plot separate graphs '
                             'for each metric, or for each type',
                             default='metric')
    plot_parser.add_argument('-l', '--log', action='store_true',
                             help='By default graphs are plot with both axes '
                             'in linear scale. This option enables plotting '
                             'the x-axis in log scale instead')

    auc_parser = subparsers.add_parser('auc', help='Calculates the Area Under '
                                       'Curve (AUC) for each metric and '
                                       'advanced type in the given input file '
                                       'generated by the merge script')
    auc_parser.add_argument('comb_file', help='The input file generated by '
                            'the FLITSR merge script')
    auc_parser.add_argument('-m', '--metrics', nargs='+', choices=met_names,
                            help='Specifies the subset of metrics to '
                            'calculate AUC values for. Allowed values are: '
                            f'[{", ".join(met_names)}]', metavar='METRIC')
    auc_parser.add_argument('-c', '--cutoff', action='store', type=float,
                            help='Specifies the percentage cut-off point to '
                            'use for the AUC calculations', default='101.0')
    argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    if (args.mode == "combine"):
        infile = args.input_file
        lines = open(infile).readlines()
    elif (args.mode == "plot"):
        plot(args.plot_file, log=args.log, all=args.all, type=args.split_type,
             metrics=args.metrics, flitsrs=args.types)
    elif (args.mode == "auc"):
        comb_file = args.comb_file
        metrics, modes, points = read_comb_file(comb_file)
        cutoff = args.cutoff
        if (args.metrics):
            metrics = args.metrics
        for metric in metrics:
            base = auc_calc(points[(metric, "Base metric")], cutoff)
            flitsr = auc_calc(points[(metric, "FLITSR")], cutoff)
            flitsr_m = auc_calc(points[(metric, "FLITSR*")], cutoff)
            print("{} FLITSR imprv: {:.6%}".format(metric, flitsr/base))
            print("{} FLITSR* imprv: {:.6%}".format(metric, flitsr_m/base))


if __name__ == "__main__":
    main()



import math
import argparse
import sys
from typing import List
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking, Tiebrk


class Suspicious():
    """
    An implementation of ranking metric used for
    fault localization
    """
    inf = 2**48

    def __init__(self, ef: int, tf: int, ep: int, tp: int):
        """
        The four basic counts that are parameters to
        different metrics
        :param ef : no. of failed tests that use e
        :param tf : total failed tests
        :param ep : no. of passed tests that use e
        :param tp : total passed tests
        """
        self.ef = ef
        self.tf = tf
        self.ep = ep
        self.tp = tp
        self.nf = tf - ef
        self.np = tp - ep

    def execute(self, metric: str) -> float:
        func = getattr(self, metric)
        return func()

    @staticmethod
    def apply_formula(spec: Spectrum, formula: str,
                      tiebrk: Tiebrk, reverse=True) -> Ranking:
        """
        Calculate the scores for each of the elements using the given formula.
        Assumes a non-empty spectrum.
        """
        ranking: Ranking = Ranking(tiebrk)
        for elem in spec.groups():
            sus = Suspicious(spec.f[elem], spec.tf, spec.p[elem], spec.tp)
            exect = spec.p[elem]+spec.f[elem]
            ranking.append(elem, sus.execute(formula), exect)
        ranking.sort(reverse)
        return ranking

    @staticmethod
    def getNames(all_names=False) -> List[str]:
        if (all_names):
            all_names = dir(Suspicious)
            names = [x for x in all_names if (not x.startswith("_")
                     and x != "execute" and x != "getNames"
                     and x != "apply_formula")]
        else:
            names = ['artemis', 'barinel', 'dstar', 'gp13', 'harmonic',
                     'hyperbolic', 'jaccard', 'naish2', 'ochiai', 'overlap',
                     'parallel', 'tarantula', 'zoltar']
        return names

    def ample(self) -> float:
        t1 = 0.0 if (self.ef == 0) else self.ef/self.tf
        t2 = 0.0 if (self.ep == 0) else self.ep/self.tp
        return abs(t1 - t2)

    def anderberg(self) -> float:
        if (self.ef == 0):
            return 0.0
        denominator = (self.ef + 2*(self.nf + self.ep))
        return self.ef/denominator

    def arith_mean(self) -> float:
        denominator = (self.ef + self.ep)*(self.np + self.nf) + self.tf*self.tp
        nominator = 2*self.ef*self.np - 2*self.nf*self.ep
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def artemis(self) -> float:
        """Dummy implementation to add artemis to list of available metrics"""
        raise NotImplementedError("Artemis not implemented as a basic metric")

    def cohen(self) -> float:
        denominator = (self.ef + self.ep)*self.tp + self.tf*(self.nf + self.np)
        nominator = 2*self.ef*self.np - 2*self.nf*self.ep
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def dice(self) -> float:
        if (self.ef == 0):
            return 0.0
        denominator = self.tf + self.ep
        return 2*self.ef/denominator

    def euclid(self) -> float:
        return math.sqrt(self.ef + self.np)

    def fleiss(self) -> float:
        denominator = (2*self.ef*self.nf*self.ep) + (2*self.np*self.nf*self.ep)
        nominator = (4*self.ef*self.np) - (4*self.nf*self.ep) - (self.nf - self.ep)**2
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def geometric(self) -> float:
        denominator = math.sqrt((self.ef + self.ep)*(self.np + self.nf)*self.tf*self.tp)
        nominator = self.ef*self.np - self.nf*self.ep
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def goodman(self) -> float:
        denominator = 2*self.ef + self.nf + self.ep
        nominator = 2*self.ef - self.nf - self.ep
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def hamann(self) -> float:
        denominator = self.tf + self.tp
        nominator = self.ef + self.np  - self.nf - self.ep
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def hamming(self) -> float:
        return self.ef + self.np

    def harmonic(self) -> float:
        n1 = (self.ef*self.np - self.nf*self.ep)
        n2 = ((self.ef + self.ep)*(self.np + self.nf) + self.tf*self.tp)
        nominator = n1*n2
        denominator = (self.ef + self.ep)*(self.np + self.nf)*self.tf*self.tp
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def jaccard(self) -> float:
        """
        Ref: Chen, M. Y., Kiciman, E., Fratkin, E., Fox, A., and Brewer, E. A.
        Pinpoint: Problem determination in large, dynamicinternet services.
        In 2002 International Conference on Dependable Systems
        and Networks (DSN 2002), 23-26 June 2002,
        Bethesda, MD, USA, Proceedings(2002), IEEE Computer Society,
        pp. 595–604
        """
        if self.ef == 0 :
            return 0.0
        denominator = self.tf + self.ep
        return self.ef/denominator

    def kulczynski1(self) -> float:
        denominator = self.nf + self.ep
        if (self.ef == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return self.ef/denominator

    def kulczynski2(self) -> float:
        t1 = 0.0 if (self.ef == 0) else self.ef/self.tf
        t2 = 0.0 if (self.ef == 0) else self.ef/(self.ef + self.ep)
        return 0.5*(t1 + t2)

    def m1(self) -> float:
        denominator = self.nf + self.ep
        nominator = self.ef + self.np
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def m2(self) -> float:
        if (self.ef == 0):
            return 0.0
        denominator = self.ef + self.np + 2*(self.ef + self.ep)
        return self.ef/denominator

    def ochiai(self) -> float:
        """
        Ref: Ochiai, A. Zoogeographical studies on the soleoid
        fishes found in japan and its neighhouring regions-ii.
        Bulletin of the Japanese Society of Scientific Fisheries
        22, 9 (1957), 526–530
        """
        denominator = math.sqrt(self.tf * (self.ef + self.ep))
        if (self.ef == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return self.ef/denominator

    def ochiai2(self) -> float:
        nominator = self.ef*self.np
        denominator = math.sqrt((self.ef + self.ep)*(self.np + self.nf)*self.tf*self.tp)
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def overlap(self) -> float:
        denominator = min(self.ef, self.nf, self.ep)
        if (self.ef == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return self.ef/denominator

    def parallel(self) -> float:
        """Dummy implementation to add parallel to list of available metrics"""
        raise NotImplementedError("Parallel not implemented as a basic metric")

    def rogers_tanimoto(self) -> float:
        nominator = self.ef + self.np
        denominator = self.ef + self.np + 2*(self.nf + self.ep)
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def rogot1(self) -> float:
        t1 = 0.0 if (self.ef == 0) else self.ef/(2*self.ef + self.nf + self.ep)
        t2 = 0.0 if (self.np == 0) else self.np/(2*self.np + self.nf + self.ep)
        return 0.5*(t1 + t2)

    def rogot2(self) -> float:
        t1 = 0.0 if (self.ef == 0) else self.ef/(self.ef + self.ep)
        t2 = 0.0 if (self.ef == 0) else self.ef/self.tf
        t3 = 0.0 if (self.np == 0) else self.np/self.tp
        t4 = 0.0 if (self.np == 0) else self.np/(self.np + self.nf)
        return 0.25*(t1 + t2 + t3 + t4)

    def russell_rao(self) -> float:
        if (self.ef == 0):
            return 0.0
        denominator = self.tf + self.tp
        return self.ef/denominator

    def sbi(self) -> float:
        if (self.ef == 0):
            return 0.0
        return self.ef / (self.ef + self.ep)

    def scott(self) -> float:
        nominator = 4*self.ef*self.np - 4*self.nf*self.ep - (self.nf - self.ep)**2
        denominator = (2*self.ef + self.nf + self.ep)*(2*self.np + self.nf + self.ep)
        if (nominator == 0):
            return 0.0
        elif (denominator == 0):
            return Suspicious.inf
        return nominator/denominator

    def simpl_match(self) -> float:
        nominator = self.ef + self.np
        if (nominator == 0):
            return 0.0
        denominator = self.tf + self.tp
        return nominator/denominator

    def sokal(self) -> float:
        nominator = 2*(self.ef + self.np)
        denominator = nominator + self.nf + self.ep
        if (nominator == 0):
            return 0.0
        return nominator/denominator

    def sorensen_dice(self) -> float:
        nominator = 2*self.ef
        if (nominator == 0):
            return 0.0
        denominator = nominator + self.nf + self.ep
        return nominator/denominator

    def tarantula(self) -> float:
        """
        Ref: Jones, J. A., and Harrold, M. J.Empirical evaluation
        of the Tarantula automatic fault-localization technique.
        In 20th IEEE/ACM International Conference on
        Automated Software Engineering (ASE 2005),
        November 7-11, 2005, Long Beach,CA, USA(2005),
        D. F. Redmiles, T. Ellman, and A. Zisman, Eds., ACM, pp. 273–282
        """
        if (self.ef == 0):
            return  0.0
        nominator = self.ef / self.tf
        passed_component = 0.0 if (self.ep == 0) else self.ep/self.tp
        denominator = nominator + passed_component
        return nominator/denominator

    def wong1(self) -> float:
        return self.ef

    def wong2(self) -> float:
        return self.ef - self.ep

    def wong3(self) -> float:
        if (self.ep <= 2):
            return self.ef - self.ep
        elif (self.ep <= 10):
            return self.ef - (2 + 0.1*(self.ep - 2))
        else:
            return self.ef - (2.8 + 0.001*(self.ep - 10))

    def zoltar(self) -> float:
        if self.ef == 0 :
            return  0.0
        multifault_component = 10000*(self.nf*self.ep)/self.ef
        denominator =  self.tf + self.ep + multifault_component
        return self.ef / denominator

    def naish2(self) -> float:
        """
        Ref: Naish, L., Lee, H. J., Ramamohanarao, K. A model for spectra-based software
        diagnosis. ACM Trans. Softw. Eng. Methodol. (2011), 20(3):1-32
        """
        return self.ef - (self.ep/(self.tp+1))

    def dstar(self, p=2) -> float:
        """
        Ref: Wong, W. E., Debroy, V., Gao, R., and Li, Y.
        The dstar method for effective software fault localization.
        IEEE Trans. Reliability 63, 1 (2014), 290–308
        """
        nominator = self.ef**p
        denominator = self.ep + self.nf
        if (nominator == 0):
            return 0.0
        elif denominator == 0:
            return Suspicious.inf
        return nominator/denominator

    def gp13(self) -> float:
        """
        Yoo, S. Evolving human competitive spectra-based fault localisation
        techniques. In: Proceedings of the 4th International Conference onSearch
        Based Software Engineering, (2012).
        """
        if (self.ef == 0):
            return 0.0
        denominator = 2*self.ep + self.ef
        return self.ef*(1 + 1/denominator)

    def hyperbolic(self) -> float:
        if (self.ef + self.ep == 0 or self.tf == 0):
            return 0.0
        K1 = 0.375
        K2 = 0.768
        K3 = 0.711
        t1 = 1/(K1 + (self.nf/self.tf))
        t2 = K3/(K2 + self.ep/(self.ef + self.ep))
        return t1 + t2

    def _old_barinel(self) -> float:
        if (self.nf == 0 or self.ep + self.ef == 0):
            return 0.0
        if (self.ep == 0):
            h = 0.0
        else:
            h = self.ep/(self.ep + self.ef)
        return h**(self.ep) * (1-h)**(self.ef)

    def barinel(self) -> float:
        if (self.nf == 0 or self.ep + self.ef == 0):
            return 0.0
        if (self.ep == 0):
            h = 0.0
        else:
            h = self.ep/(self.ep + self.ef)
        return h**(self.ep) * (1-h)**(11)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Prints the pre-selected "
                                     "list of popular metrics.")
    parser.add_argument('-a', '--all', action='store_true', help="Print all "
                        "available metrics instead of only the pre-selected "
                        "popular metrics.")
    args = parser.parse_args(sys.argv[1:])
    names = Suspicious.getNames(args.all)
    print(*names)



import sys
from os import path as osp
from typing import List
import os

if __name__ == "__main__":
    if (len(sys.argv) < 3):
        print("USAGE: python3 stmt_to_method <version> <save dir>")
        quit()
    version = sys.argv[1]
    savedir = sys.argv[2]
    spectra_lines = open(osp.join(version,"spectra.csv")).readlines()[1:]
    lines = [line.strip().split(":") for line in spectra_lines]
    coverage_all = open(osp.join(version, "matrix.txt")).readlines()
    coverage: List[List[str]] = [[] for _ in range(len(coverage_all))]
    for i in range(len(coverage_all)):
        coverage[i] = coverage_all[i].split()
    i = 0
    new_methods = []
    collapses = []
    while (i < len(lines)):
        method = lines[i][0]
        first_line = lines[i][1]
        bugs = []
        if (len(lines[i]) > 2):
            bugs.append(lines[i][2])
        j = i
        while (j+1 < len(lines) and lines[j+1][0] == method):
            j += 1
            if (len(lines[j]) > 2):
                bugs.append(lines[j][2])
        new_methods.append([method] + [first_line] + bugs)
        collapses.append((i, j))
        i = j + 1
    new_coverage = []
    for test_cov in coverage:
        new_cov = []
        x = 0
        collect = 0
        for i in range(len(test_cov)-1):
            if (i < collapses[x][0]):
                new_cov.append(test_cov[i])
            elif (i >= collapses[x][0] and i < collapses[x][1]):
                collect = collect or int(test_cov[i])
            elif (i == collapses[x][1]):
                new_cov.append(str(collect))
                collect = 0
                x += 1
            else:
                raise Exception(str(i)+" is not between range: "+str(collapses[x]))
        new_cov.append(test_cov[-1])
        new_coverage.append(new_cov)
    os.mkdir(osp.join(savedir, version))
    spec_file = open(osp.join(savedir, version, "spectra.csv"), 'w')
    spec_file.writelines([":".join(x)+"\n" for x in new_methods])
    cov_file = open(osp.join(savedir, version, "matrix.txt"), 'w')
    cov_file.writelines([" ".join(x)+"\n" for x in new_coverage])



import random
from typing import List, Optional, Dict
from enum import Enum, auto
from flitsr.spectrum import Spectrum
import copy


class Tiebrk(Enum):
    EXEC = auto()
    RNDM = auto()
    ORIG = auto()


class Rank:
    """
    A group from a spectrum, together with its scores.
    """
    def __init__(self, group: Spectrum.Group, score: float,
                 exec_count: int):
        self.group = group
        self.score = score
        self.exec = exec_count

    def __str__(self):
        return "(" + str(self.group) + ", " + str(self.score) + ")"

    def __repr__(self):
        return str(self)


class Ranking:
    def __init__(self, tiebrk: Tiebrk = Tiebrk.ORIG):
        self._ranks: List[Rank] = []
        self._tiebrk = tiebrk
        self.group_map: Dict[Spectrum.Group, Rank] = {}
        self.place = 0

    def __getitem__(self, index: int):
        return self._ranks[index]

    def get_rank(self, group: Spectrum.Group) -> Rank:
        return self.group_map[group]

    def sort(self, reverse: bool):
        if (self._tiebrk == Tiebrk.EXEC):  # Sorted by execution counts
            self._ranks.sort(key=lambda x: x.exec, reverse=reverse)
        elif (self._tiebrk == Tiebrk.RNDM):  # random ordering
            random.shuffle(self._ranks)
        elif (self._tiebrk == Tiebrk.ORIG):  # original ranking tie break
            if (orig is not None):  # sort by original rank then exec count
                self._ranks.sort(key=lambda x: orig.get_rank(x.group).exec,
                                 reverse=reverse)
                self._ranks.sort(key=lambda x: orig.get_rank(x.group).score,
                                 reverse=reverse)
            else:  # if no orig, still sort by current execution count
                self._ranks.sort(key=lambda x: x.exec, reverse=reverse)
        self._ranks.sort(key=lambda x: x.score, reverse=reverse)

    def append(self, group: Spectrum.Group, score: float, exec_count: int):
        created = Rank(group, score, exec_count)
        self._ranks.append(created)
        self.group_map[group] = created

    def extend(self, ranks: List[Rank]):
        self._ranks.extend(ranks)
        for rank in ranks:
            self.group_map[rank.group] = rank

    def has_group(self, group: Spectrum.Group) -> bool:
        return group in self.group_map

    def __iter__(self):
        return iter(self._ranks)

    def __len__(self):
        return len(self._ranks)


orig: Optional[Ranking] = None


def set_orig(ranking: Ranking):
    global orig
    orig = copy.deepcopy(ranking)


def unset_orig():
    global orig
    orig = None



import re
from typing import Tuple, List, Union, Any
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking


class RankingSpectrumBuilder:
    def __init__(self):
        self._elements: List[Spectrum.Element] = []
        self._groups: List[Spectrum.Group] = []

    def addElement(self, details: List[str],
                   faults: List[Any]) -> Spectrum.Element:
        e = Spectrum.Element(details, len(self._elements), faults)
        self._elements.append(e)
        return e

    def createGroup(self, elems: List[Spectrum.Element]) -> Spectrum.Group:
        """
        Hard-code the groups for a pre-determined spectrum
        """
        i = len(self._groups)
        g = Spectrum.Group(elems, i)
        self._groups.append(g)
        for e in elems:
            e.set_group(g)
        return g

    def get_spectrum(self):
        return Spectrum(self._elements, self._groups, [], {})


def read_any_ranking(ranking_file: str,
                     method_level=False) -> Tuple[Ranking, Spectrum]:
    f = open(ranking_file)
    if (f.readline().startswith("Faulty grouping")):
        return read_flitsr_ranking(ranking_file)
    else:
        return read_ranking(ranking_file, method_level)


def read_ranking(ranking_file: str,
                 method_level=False) -> Tuple[Ranking, Spectrum]:
    f = open(ranking_file)
    spectrumBuilder = RankingSpectrumBuilder()
    ranking = Ranking()
    i = 0  # number of actual lines
    bugs = 0
    method_map = {}
    methods = {}
    f.readline()
    for line in f:
        line = line.strip()
        score = float(line[line.index(";")+1:])
        name = line[:line.index(";")]
        l = name.strip().split(':')
        r = re.search("(.*)\\$(.*)#([^:]*)", l[0])
        if (r is None):
            raise ValueError("Incorrectly formatted line \"" + line +
                             "\" when reading input ranking file")
        faults = []
        if (len(l) > 2):
            if (not l[2].isdigit()):
                faults = [bugs]
            else:
                faults = []
                for b in l[2:]:
                    faults.append(int(b))
            bugs += 1
        if (method_level):
            # Add the method to the spectrum
            details = [r.group(1)+"."+r.group(2), r.group(3), l[1]]
            if ((details[0], details[1]) not in methods):
                # add with first line number
                elem = spectrumBuilder.addElement(details, faults)
                spectrumBuilder.createGroup([elem])
                methods[(details[0], details[1])] = elem
                method_map[i] = elem
            else:
                elem = method_map[i] = methods[(details[0], details[1])]
                for fault in faults:
                    if (fault not in elem.faults):
                        elem.faults.append(fault)
            # Add/Update the method's score
            elem = method_map[i]
            group = elem.group()
            if (ranking.has_group(group)):
                rank_elem = ranking.get_rank(group)
                rank_elem.score = max(rank_elem.score, score)
            else:
                ranking.append(group, score, 0)
        else:
            details = [r.group(1)+"."+r.group(2), r.group(3), l[1]]
            elem = spectrumBuilder.addElement(details, faults)
            spectrumBuilder.createGroup([elem])
            method_map[i] = elem
            ranking.append(elem.group(), score, 0)
        i += 1
    spectrum = spectrumBuilder.get_spectrum()
    return ranking, spectrum


def read_flitsr_ranking(ranking_file: str) -> Tuple[Ranking, Spectrum]:
    f = open(ranking_file)
    spectrumBuilder = RankingSpectrumBuilder()
    ranking = Ranking()
    num_locs = 0  # number of reported locations (methods/lines)
    i = 0  # number of actual lines
    line = f.readline()
    while (line != ""):
        line = line.strip()
        score: Union[int, float]
        str_score = line[line.index(": ")+2:line.index(" [")]
        if (str_score.isdigit()):
            score = int(str_score)
        else:
            score = float(str_score)
        line = f.readline().strip()
        group_elems = []
        while (not line.startswith("]")):
            # read in ranked element (old or new FLITSR format)
            m = re.fullmatch("\\s*(?:\\([0-9]+\\)\\s*)?(\\S*)\\s*(?:\\(FAULT ([0-9.,]+)\\))?", line)
            if (m is None):
                # if normal format fails, try DUA format
                m = re.fullmatch("\\s*(?:\\([0-9]+\\)\\s*)?(\\S*\\s\\S*\\s\\S*)\\s*(?:\\(FAULT ([0-9.,]+)\\))?", line)
                if (m is None):
                    raise ValueError("Incorrectly formatted line \"" + line +
                                     "\" when reading input ranking file")
            details = m.group(1).split('|')
            if (m.group(2)):
                faults = [int(i) if i.isdecimal() else float(i)
                          for i in m.group(2).split(',')]
            else:
                faults = []
            elem = spectrumBuilder.addElement(details, faults)
            group_elems.append(elem)
            i += 1
            line = f.readline().strip()
        group = spectrumBuilder.createGroup(group_elems)
        ranking.append(group, score, 0)
        num_locs += 1
        line = f.readline().strip()
    spectrum = spectrumBuilder.get_spectrum()
    return ranking, spectrum



from __future__ import annotations
import pkgutil
from flitsr import advanced
from enum import Enum
import importlib
from typing import Union
import sys

_rankers = {}
_clusters = {}
_refiners = {}
all_types = {}
all_types_print = []


def register_ranker(cls):
    if (cls.__name__.upper() in all_types):
        raise ValueError(f'Duplicate advanced types for name "{cls.__name__}"')
    _rankers[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


def register_cluster(cls):
    if (cls.__name__.upper() in all_types):
        raise ValueError(f'Duplicate advanced types for name "{cls.__name__}"')
    _clusters[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


def register_refiner(cls):
    if (cls.__name__.upper() in all_types):
        raise ValueError(f'Duplicate advanced types for name "{cls.__name__}"')
    _refiners[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


# load local advanced types
__all__ = [m[1] for m in pkgutil.iter_modules(advanced.__path__)]
for module in __all__:
    importlib.import_module('.'+module, package=__name__)
# load plugin advanced types
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points
adv_entry_points = entry_points(group='flitsr.advanced')
for adv_ep in adv_entry_points:
    adv_ep.load()

RefinerType = Enum('RefinerType', _refiners,  # type:ignore
                   module=advanced, qualname='advanced.RefinerType')
ClusterType = Enum('ClusterType', _clusters,  # type:ignore
                   module=advanced, qualname='advanced.ClusterType')
RankerType = Enum('RankerType', _rankers,  # type:ignore
                  module=advanced, qualname='advanced.RankerType')

for cls in all_types.values():
    if (hasattr(cls, '__print_name__')):
        all_types_print.append(cls.__print_name__.lower())
    else:
        all_types_print.append(cls.__name__.lower())


class Config:
    def __init__(self, ranker: RankerType = None, cluster: ClusterType = None,
                 refiner: RefinerType = None):
        self.cluster = cluster
        self.ranker = ranker
        self.refiner = refiner

    def __str__(self):
        return self._get_str()

    def _get_str(self, printed=False):
        c = '_' if printed else '+'
        components = [self.ranker or RankerType['SBFL'], self.refiner,
                      self.cluster]
        string = c.join(self._get_name(c, printed) for c in components
                        if c is not None)
        return string

    def _get_name(self, typ: Union[RefinerType, ClusterType, RankerType],
                  printed) -> str:
        if (printed and hasattr(typ.value, '__print_name__')):
            return typ.value.__print_name__.lower()
        elif (printed):
            return typ.name.lower()
        else:
            return typ.name.upper()

    def __repr__(self):
        return self._get_str()

    def get_file_name(self):
        return self._get_str(printed=True)



from typing import Collection, Any


def existing(*params: str):
    """
    A python decorator to mark the given parameter(s) as one(s) already defined
    in FLITSR's arguments. These parameters for the decorated advanced type are
    therefore not added as FLITSR command line arguments, but are instead taken
    from the existing command line arguments.
    """
    def add_existing(fn):
        if (not hasattr(fn, '__existing__')):
            setattr(fn, '__existing__', list())
        existing = getattr(fn, '__existing__')
        for param in params:
            existing.append(param)
        return fn
    return add_existing


def choices(param: str, choices: Collection[Any]):
    """
    An advanced types decorator to specfiy the choices available for a given
    parameter. The given choices will be added as available choices for the
    command line argument for FLITSR.
    """
    def add_choices(fn):
        if (not hasattr(fn, '__choices__')):
            setattr(fn, '__choices__', dict())
        choices_dict = getattr(fn, '__choices__')
        choices_dict[param] = choices
        return fn
    return add_choices


def print_name(print_name: str):
    """
    An advanced types decorator to specfiy an alternative name for the
    decorated class to use when printing results to a file. By default the
    lower-case name of the advanced type is used.
    """
    def add_print_name(cls):
        setattr(cls, '__print_name__', print_name)
        return cls
    return add_print_name



import importlib
import pkgutil
from enum import Enum
from flitsr import input

_inputs = {}


def register_input(cls):
    _inputs[cls.__name__.upper()] = cls


__all__ = [m[1] for m in pkgutil.iter_modules(input.__path__)]
for module in __all__:
    importlib.import_module('.'+module, package=__name__)

InputType = Enum('InputType', _inputs,  # type:ignore
                 module=input, qualname='input.InputType')



import sys
import os
from os import path as osp
from io import TextIOWrapper
import re
import mmap
from typing import Dict, Tuple, List
from flitsr.split_faults import split
from flitsr.spectrum import Spectrum, Outcome
from flitsr.spectrumBuilder import SpectrumBuilder
from flitsr import errors
from flitsr.input.input_reader import Input


class TCM(Input):
    def get_run_file_name(self, input_path: str):
        return re.sub("\\.\\w+$", ".run", input_path)

    def construct_details(self, f: TextIOWrapper, method_level: bool,
                          sb: SpectrumBuilder) -> Dict[int, Spectrum.Element]:
        """
        Fills the spectrum object with elements read in from the open file 'f'.
        """
        i = 0  # number of actual lines
        method_map: Dict[int, Spectrum.Element] = {}
        methods: Dict[Tuple[str, str], Spectrum.Element] = {}
        bugs = 0
        line = f.readline()
        while (not line == '\n'):
            m = re.fullmatch('((?:[^ ]| [^|])+)(?: \\| (.+))?', line.rstrip())
            if (m is None):
                errors.error("incorrectly formatted element line in input file:",
                             line, "terminating...")
            else:
                faults = []
                if (m.group(2) is not None):
                    bs = m.group(2).split(' | ')
                    if (not bs[0].isdigit()):
                        faults = [bugs]
                    else:
                        faults = []
                        for b in bs:
                            faults.append(int(b))
                    bugs += 1
                if (method_level):
                    details_m = re.match('([^:]+):([^:]+):([0-9]+)', m.group(1))
                    if (details_m is None):
                        errors.error("Incorrectly formatted line \"" + line +
                                     "\" for method level evaluation")
                    else:
                        details = list(details_m.groups())
                    if ((details[0], details[1]) not in methods):
                        # append with first line number
                        elem = sb.addElement(details, faults)
                        methods[(details[0], details[1])] = elem
                        method_map[i] = elem
                    else:
                        elem = method_map[i] = methods[(details[0], details[1])]
                        # update method's faults
                        for fault in faults:
                            if (fault not in elem.faults):
                                elem.faults.append(fault)
                else:
                    details = m.group(1).split(":")
                    elem = sb.addElement(details, faults)
                    method_map[i] = elem
                i += 1
                line = f.readline()
        return method_map

    def construct_tests(self, f: TextIOWrapper, sb: SpectrumBuilder):
        line = f.readline()
        i = 0
        while (not line == '\n'):
            m = re.fullmatch('([^ ]+) (PASSED|FAILED|ERROR)( .*)?', line.rstrip())
            if (m is None):
                errors.error("incorrectly formatted test line in input file:",
                             line, "terminating...")
            else:
                sb.addTest(m.group(1), i, Outcome[m.group(2)])
            line = f.readline()
            i += 1

    def fill_spectrum(self, f: TextIOWrapper,
                      method_map: Dict[int, Spectrum.Element],
                      sb: SpectrumBuilder):
        for t, test in enumerate(sb.get_tests()):
            line = f.readline()
            if (line == ''):
                errors.error('Incorrect number of matrix lines', f'({t})',
                             'in input file, terminating...')
            arr = line.rstrip().split(' ')
            seen = set()
            for i in range(0, int(len(arr)/2)):
                try:
                    elem = method_map[int(arr[i*2])]
                except KeyError:
                    errors.error(f'Incorrect element reference "{arr[i*2]}"',
                                 f'at column {i*4+1} on line {t} of the matrix',
                                 'component in the input file, terminating...')
                if (elem not in seen):
                    sb.addExecution(test, elem)
                    seen.add(elem)

    def read_spectrum(self, input_path: str, split_faults: bool,
                      method_level=False) -> Spectrum:
        sb = SpectrumBuilder()
        method_map: Dict[int, Spectrum.Element]
        file = open(input_path)
        exec_checks = {'#tests': False, '#uuts': False, '#matrix': False}
        while (True):
            line = file.readline()
            if (line == '' or not line.startswith("#")):
                break
            elif (line.startswith("#metadata")):
                while (not line == '\n'):
                    line = file.readline()
            elif (line.startswith("#tests")):
                if (exec_checks['#tests'] is True):
                    print("WARNING: duplicate #tests component in input file",
                          file=sys.stderr)
                # Constructing the spectrum
                self.construct_tests(file, sb)
                exec_checks['#tests'] = True
            elif (line.startswith("#uuts")):
                if (exec_checks['#uuts'] is True):
                    print("WARNING: duplicate #uuts component in input file",
                          file=sys.stderr)
                # Getting the details of the project
                method_map = self.construct_details(file, method_level, sb)
                exec_checks['#uuts'] = True
            elif (line.startswith("#matrix")):
                if (exec_checks['#matrix'] is True):
                    print("WARNING: duplicate #matrix component in input file",
                          file=sys.stderr)
                # Filling the spectrum
                self.fill_spectrum(file, method_map, sb)
                exec_checks['#matrix'] = True
            else:
                print("ERROR: Incorrectly formatted input file at line:", line,
                      "terminating...", file=sys.stderr)
                quit()
        file.close()
        if (sum(exec_checks.values()) != 3):
            missing = [e[0] for e in exec_checks.items() if e[1] is False]
            errors.error(f"Input file missing components: {missing},",
                         "terminating...")
        spectrum = sb.get_spectrum()
        del sb
        # Split fault groups if necessary
        if (split_faults):
            faults, unexposed = split(spectrum.get_faults(), spectrum)
            for elem in unexposed:
                elem.faults.clear()
                errors.warning(f"Dropped faulty UUT: {elem} due to unexposure")
            # Get element's fault lists
            fault_lists: Dict[Spectrum.Element, List[float]] = {}
            for (f_num, f_locs) in faults.items():
                for elem in f_locs:
                    fault_lists.setdefault(elem, []).append(f_num)
            # Set element's fault lists
            for (elem, f_list) in fault_lists.items():
                elem.faults = f_list
            if (len(faults) == 0):
                errors.error(f"No exposable faults in {input_path},",
                             "terminating...")
        return spectrum

    @staticmethod
    def check_format(input_path: str) -> bool:
        if (osp.isfile(input_path)):
            with open(input_path, 'rb', 0) as file:
                with mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as f:
                    return f.find(b'#uuts\n') != -1 and f.find(b'#tests\n') != -1 \
                        and f.find(b'#matrix\n') != -1
        return False

    @classmethod
    def write_spectrum(cls, spectrum: Spectrum, output_path: str):
        """ Output the spectrum in TCM format """
        with open(output_path, 'w') as file:
            type_ = cls.get_type()
            print("#tests", file=file)
            for test in spectrum.tests():
                print(test.name, test.outcome.name, file=file)
            print(file=file)
            print("#uuts", file=file)
            # TODO: change _elements below to elements()
            for elem in spectrum._elements:
                print(elem.output_str(type_=type_), file=file)
            print(file=file)
            print("#matrix", file=file)
            for test in spectrum.tests():
                first = True
                for elem_id, elem in enumerate(spectrum._elements):
                    if (spectrum[test][elem]):
                        print(("" if first else " ")+str(elem_id), "1", end="",
                              file=file)
                        first = False
                print(file=file)

    @classmethod
    def get_elem_separators(cls):
        return ['.', ':', ':', ' | ']


if __name__ == "__main__":
    from flitsr.output import print_spectrum
    d = sys.argv[1]
    tinput = TCM()
    spectrum = tinput.read_spectrum(d, False)
    print_spectrum(spectrum)



import subprocess
import re
import copy
import os
from typing import List
import tempfile
from importlib import resources
from flitsr.spectrum import Spectrum
from flitsr.advanced.cluster import Cluster
from flitsr.advanced.attributes import choices
from flitsr.input import InputType


class Parallel(Cluster):
    """
    Run one of the parallel debugging algorithms on the spectrum to
    produce multiple spectrums, and process all other options on
    each spectrum.
    """

    @choices('parType', ['bdm', 'msp', 'hwk', 'vwk'])
    def __init__(self, parType: str = 'msp'):
        self.parType = parType

    def cluster(self, inp_file: str, spectrum: Spectrum,
                method_lvl=False) -> List[Spectrum]:
        # write the spectrum to a temporary file
        tmp_fd, tmp_name = tempfile.mkstemp(text=True)
        inp_file = tmp_name
        InputType['TCM'].value.write_spectrum(spectrum, tmp_name)
        # Run the parallel algorithm
        spectrums = self.partition_table(inp_file, spectrum)
        # remove temporary file
        os.remove(tmp_name)
        return spectrums

    def partition_table(self, d: str, spectrum: Spectrum) -> List[Spectrum]:
        jar_name = 'parallel-1.0-SNAPSHOT-jar-with-dependencies.jar'
        with resources.path('flitsr.advanced', jar_name) as jar_file:
            output = subprocess.check_output(['java', '-jar', str(jar_file), d,
                                              self.parType]).decode('utf-8')
        partitions = re.split("partition \\d+\n", output)[1:]
        spectrums: List[Spectrum] = []
        for partition in partitions:
            new_spectrum = copy.deepcopy(spectrum)
            tests = [int(i) for i in partition.strip().split("\n")]
            toRemove = set()
            all_tests = spectrum.tests()
            # keep failing + all passing
            for test in spectrum.failing():
                ind = all_tests.index(test)
                if (ind not in tests):
                    toRemove.add(test)
            for test in toRemove:
                new_spectrum.remove(test, hard=True)
            self.trim_groups(new_spectrum)
            spectrums.append(new_spectrum)
        return spectrums

    def trim_groups(self, spectrum):
        """Remove groups that are not in this block (i.e. ef = 0)"""
        toRemove = set()
        for group in spectrum.groups():
            if (spectrum.f[group] == 0):
                toRemove.add(group)
        for group in toRemove:
            spectrum.remove_group(group)



from __future__ import annotations
from typing import List, Dict, Any, Set, Sequence, Tuple, Callable, TYPE_CHECKING
from bitarray import bitarray
import numpy as np
from enum import Enum
from abc import ABC, abstractmethod
if TYPE_CHECKING:
    from flitsr.input import InputType

class Outcome(Enum):
    PASSED = 0
    PASS = 0
    FAILED = 1
    FAIL = 1
    ERROR = 2


class Spectrum:
    """An implementation for a program spectrum."""
    class Test:
        """A test object holds information pertaining to a particular test."""

        __slots__ = ('name', 'index', 'outcome')

        def __init__(self, name: str, index: int, outcome: Outcome):
            self.name = name
            self.index = index
            self.outcome = outcome

        def __str__(self):
            return self.name

        def __repr__(self):
            return str(self)

        def __eq__(self, other):
            return self.name == other.name

        def __hash__(self):
            return hash(self.name)

    class Entity(ABC):
        @abstractmethod
        def isFaulty(self) -> bool:
            pass

        @abstractmethod
        def index(self) -> int:
            pass

    class Element(Entity):
        """
        An element object holds information pertaining to a single spectral
        element (line, method, class, etc...).
        """
        __slots__ = ('_index', '_group', 'path', 'method', 'line', 'faults',
                     'tup')

        def __init__(self, details: List[str], index: int, faults: List[Any],
                     group: Spectrum.Group = None):
            if (len(details) < 1):
                raise ValueError("Unnamed element: ", *details)
            self._index = index
            self._group = group
            self.path = details[0]
            self.method = None
            self.line = None
            if (len(details) > 2):
                if (details[2].isdigit()):
                    self.line = int(details[2])
                    self.method = details[1]
                else:
                    self.method = details[1] + "|" + details[2]
            elif (len(details) > 1):
                if (not details[1].isdigit()):
                    self.method = details[1]
                else:
                    self.line = int(details[1])
            self.faults = faults
            self.tup = (self.path, self.method, self.line)
            # self.hash = hash(self.tup)

        def isFaulty(self) -> bool:
            return len(self.faults) > 0

        def index(self) -> int:
            return self._index

        def set_group(self, group: Spectrum.Group):
            self._group = group

        def group(self) -> Spectrum.Group:
            if (self._group is None):
                raise UnboundLocalError("Element group has not been set")
            return self._group

        def __str__(self):
            return "|".join(str(i) for i in self.tup if i) + \
                   (" (FAULT {})".format(",".join(str(x) for x in self.faults))
                    if self.faults else "")

        def output_str(self, type_: 'InputType', incl_faults=True) -> str:
            seps = type_.value.get_elem_separators()
            gstring = ''
            path_part = self.path.rpartition('.')
            if (path_part[0] != '' and path_part[2] != ''):
                gstring = path_part[0] + seps[0] + path_part[2]
            elif (path_part[0] != '' or path_part[2] != ''):
                gstring = path_part[0] + path_part[2]
            if (self.method):
                gstring += (seps[1] if (gstring != '') else '') + self.method
            if (self.line):
                gstring += (seps[2] if (gstring != '') else '') + str(self.line)
            if (incl_faults and self.isFaulty()):
                gstring += seps[3] + seps[3].join(str(x) for x in self.faults)
            return gstring

        def __repr__(self):
            return str(self)

        def __eq__(self, other):
            return self._index == other._index

        def __hash__(self):
            return self._index

        def semantic_eq(self, other):
            return self.tup == other.tup

    class Group(Entity):
        """
        A spectral group is a collection of spectral elements which have
        identical spectra.
        """
        __slots__ = ('_elems', '_is_faulty', '_index')

        def __init__(self, elems: List[Spectrum.Element] = None,
                     index: int = None):
            """
            Create a group object. If elems is given, the group is initialized
            with the elements within, otherwise an empty group is initialized.
            """
            self._elems = elems if elems is not None else []
            self._is_faulty = any(e.isFaulty() for e in self._elems)
            self._index = index

        def sort_elems(self, key: Callable):
            self._elems.sort(key=key)

        def set_index(self, index: int):
            self._index = index

        def index(self) -> int:
            if (self._index is None):
                raise UnboundLocalError("Group index has not been set")
            return self._index

        def isFaulty(self) -> bool:
            return self._is_faulty

        def get_elements(self) -> List[Spectrum.Element]:
            return self._elems

        def is_in(self, element):
            return element in self._elems

        def __str__(self) -> str:
            return f"G{self._index} ({self._elems})"

        def __repr__(self) -> str:
            return str(self)

        def __eq__(self, other):
            return self._index == other._index

        def __hash__(self):
            return self._index

    class Execution():
        """
        The Execution object holds all of the spectral information pertaining
        to the execution of a particular test.
        """
        __slots__ = ('_groups', 'exec', 'test', '_iter')

        def __init__(self, test: Spectrum.Test,
                     groups: List[Spectrum.Group]):
            self._groups = groups
            self.exec = bitarray(len(self._groups))
            self.test = test

        def update(self, group: Spectrum.Group, executed: bool):
            """
            Update the execution information of a particular group
            """
            self.exec[group.index()] = executed

        def __len__(self):
            return len(self._groups)

        def __iter__(self):
            self._iter = iter(self._groups)
            return self

        def __next__(self):
            return self.get(next(self._iter))

        def __getitem__(self, elem: Spectrum.Entity) -> bool:
            if (isinstance(elem, Spectrum.Group)):
                return self.get(elem, False)
            elif (isinstance(elem, Spectrum.Element)):
                return self.element_get(elem, False)
            else:
                raise TypeError(f"Illegal Entity type: {type(elem).__name__}")

        def __setitem__(self, group: Spectrum.Group, val: bool):
            try:
                self.exec[group.index()] = val
            except KeyError:
                pass

        def get(self, group: Spectrum.Group, default: bool = False) -> bool:
            """
            Return the execution information for the given ambiguity group. If
            the element is not present in the execution, default is given.
            """
            try:
                return bool(self.exec[group.index()])
            except (KeyError, IndexError):
                return default

        def element_get(self, elem: Spectrum.Element,
                        default: bool = False) -> bool:
            """
            Return the execution information for the given element. If the
            element is not present in the execution, default is given.
            """
            # Get the element's group
            try:
                return bool(self.exec[elem.group().index()])
            except (KeyError, IndexError):
                return default

    def __init__(self, elements: List[Spectrum.Element],
                 groups: List[Spectrum.Group], tests: List[Spectrum.Test],
                 executions: Dict[Test, Set[Element]]):
        self.spectrum: Dict[Spectrum.Test, Spectrum.Execution] = {}
        self.p: Dict[Spectrum.Group, int] = dict()
        self.f: Dict[Spectrum.Group, int] = dict()
        # Initialize element related properties
        edict = {e: i for i, e in enumerate(elements)}
        for group in groups:
            group.sort_elems(key=lambda x: edict[x])
        self._groups = sorted(groups, key=lambda g: edict[g.get_elements()[0]])
        for i, group in enumerate(self._groups):
            group.set_index(i)
            self.p[group] = 0
            self.f[group] = 0
            for elem in group.get_elements():
                elem.set_group(group)
        self._elements: List[Spectrum.Element] = elements
        # Initialize test related properties
        self._removed: List[Spectrum.Test] = []
        self._tests: List[Spectrum.Test] = tests
        self._failing: List[Spectrum.Test] = []
        self.tp: int = 0
        self.tf: int = 0
        for test in tests:
            # Increment total counts & add to failing set
            if (test.outcome is Outcome.PASSED):
                self.tp += 1
            else:
                self._failing.append(test)
                self.tf += 1
            # Add test execution
            self.spectrum[test] = self.Execution(test, self._groups)
        # Initialize execution information
        for test in self._tests:
            test_exe = executions[test]
            seen: Set[Spectrum.Group] = set()
            for elem in test_exe:
                group = elem.group()
                if (group not in seen):
                    seen.add(group)
                    self.spectrum[test].update(group, True)
                    if (test.outcome is Outcome.PASSED):
                        self.p[group] += 1
                    else:
                        self.f[group] += 1

    def __getitem__(self, t: Test):
        return self.spectrum[t]

    def __iter__(self):
        self._iter = iter(self._tests)
        return self

    def __next__(self):
        return self.spectrum[next(self._iter)]

    def __len__(self):
        return len(self._tests)

    def elements(self) -> List[Spectrum.Element]:
        return self._elements

    def groups(self) -> List[Spectrum.Group]:
        return self._groups

    def tests(self) -> List[Spectrum.Test]:
        return self._tests

    def failing(self) -> List[Spectrum.Test]:
        return self._failing

    def locs(self) -> int:
        return len(self._groups)

    def remove(self, test: Spectrum.Test, hard=False):
        self._tests.remove(test)
        if (test.outcome is Outcome.PASSED):
            self.tp -= 1
        else:
            self._failing.remove(test)
            self.tf -= 1
        for group in self.groups():
            self.remove_execution(test, group, hard=False)
        if (not hard):
            self._removed.append(test)

    def remove_execution(self, test: Spectrum.Test, ent: Spectrum.Entity,
                         hard=True):
        if (self.spectrum[test][ent]):
            if (isinstance(ent, Spectrum.Element)):
                group = ent.group()
            elif (isinstance(ent, Spectrum.Group)):
                group = ent
            if (hard):
                self.spectrum[test][group] = False
            if (test.outcome is Outcome.PASSED):
                self.p[group] -= 1
            else:
                self.f[group] -= 1

    def remove_group(self, group: Spectrum.Group):
        try:
            self._groups.remove(group)  # remove if there
        except ValueError:
            # Ignore if the element is not in the list
            pass
        self.p.pop(group, None)  # remove if there
        self.f.pop(group, None)  # remove if there
        # No need to remove execution (bitarray)
        # for test in self.tests():
        #     self.spectrum[test].remove_exec(rem)  # remove if there
        return group

    def reset(self):
        """Re-activates all the tests and recomputes counts"""
        for test in self._removed:
            self._tests.append(test)
            if (test.outcome is Outcome.PASSED):
                self.tp += 1
            else:
                self._failing.append(test)
                self.tf += 1
            for group in self.groups():
                if (self.spectrum[test][group]):
                    if (test.outcome is Outcome.PASSED):
                        self.p[group] += 1
                    else:
                        self.f[group] += 1
        self._removed.clear()

    def get_group(self, elem: Spectrum.Element) -> Spectrum.Group:
        """
        Given an element, return the group of elements with identical coverage
        that this element is apart of.
        """
        return elem.group()

    def get_faults(self) -> Dict[Any, List[Spectrum.Element]]:
        actual_faults: Dict[Any, List[Spectrum.Element]] = dict()
        for group in self.groups():
            for elem in group.get_elements():
                if (elem.faults):
                    for fault in elem.faults:
                        if (fault not in actual_faults):
                            actual_faults[fault] = []
                        actual_faults[fault].append(elem)
        return actual_faults

    def get_tests(self, entity: Spectrum.Entity, only_failing=False,
                  remove=False) -> Set[Spectrum.Test]:
        """
        Finds all the test cases executing the given element, and (optionally)
        removes them from the spectrum.
        """
        executing = set()
        tests = self.failing() if only_failing else self.tests()
        for test in tests:
            if (self[test][entity]):
                executing.add(test)
        if (remove):
            for test in executing:
                self.remove(test)
        return executing

    def get_executed_groups(self, test: Spectrum.Test) -> Set[Spectrum.Group]:
        return self._get_executed_entities(test, groups=True)

    def get_executed_elements(self, test: Spectrum.Test) -> Set[Spectrum.Element]:
        return self._get_executed_entities(test, groups=False)

    def _get_executed_entities(self, test: Spectrum.Test, groups=True):
        """
        Finds either the groups (default) or the elements (groups = False)
        executed in the given test.
        """
        executed = set()
        entities: Sequence[Spectrum.Entity]
        if (groups):
            entities = self._groups
        else:
            entities = self._elements
        for ent in entities:
            if (self[test][ent]):
                executed.add(ent)
        return executed

    def to_matrix(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Converts the current spectrum into a numpy matrix and error vector.
        """
        # If no matrix already, create one
        if (not hasattr(self, '_matrix')):
            # Use all test cases and most elements
            self._matrix_tests = np.array(list(self.spectrum.keys()))
            self._matrix_elems = np.array(list(self._groups))
            self._matrix = np.zeros((len(self._matrix_tests),
                                     len(self._matrix_elems)),
                                    dtype=bool)
            self._errVector = np.zeros(len(self._matrix_tests),
                                       dtype=bool)
            for (i, test) in enumerate(self._matrix_tests):
                for (j, elem) in enumerate(self._matrix_elems):
                    self._matrix[i][j] = self.spectrum[test][elem]
                self._errVector[i] = 0 if (test.outcome is
                                           Outcome.PASSED) else 1
        # Extract submatrix
        test_set = set(self._tests)
        tmask = np.array([item in test_set for item in self._matrix_tests])
        # tmask = np.isin(self._matrix_tests, self._tests)
        group_set = set(self._groups)
        emask = np.array([item in group_set for item in self._matrix_elems])
        # emask = np.isin(self._matrix_elems, self._groups)
        matrix = self._matrix[np.ix_(tmask, emask)]
        errVector = self._errVector[tmask]
        return matrix, errVector

    def search_tests(self, name_part, incl_removed=False):
        """
        Searches for any test that matches the partial test name `name_part`.
        Returns a list of all the matches.
        """
        results = [t for t in self.tests() if t.name.find(name_part) != -1]
        if (incl_removed):
            results.extend([t for t in self._removed
                            if t.name.find(name_part) != -1])
        return results

    def search_elements(self, name_part, groups=False):
        """
        Searches for any element that matches the partial element name
        `name_part`. Returns a list of all the matches.
        """
        elems = self._groups if groups else self._elements
        results = [e for e in elems if str(e).find(name_part) != -1]
        return results



from __future__ import annotations
import argparse
import argcomplete
import inspect
import sys
from pathlib import Path
from os import path as osp
from typing import List, Dict, Any, Optional, IO, BinaryIO, Tuple
from flitsr.suspicious import Suspicious
from flitsr import cutoff_points
from flitsr.singleton import SingletonMeta
from flitsr.ranking import Tiebrk
from flitsr import advanced
from flitsr.advanced import Config


class Args(argparse.Namespace, metaclass=SingletonMeta):
    def __init__(self):
        self._advanced_params: Dict[str, List[str]] = {}

    def _add_params(self, params: argparse.Namespace):
        dict_ = vars(params)
        for key in dict_:
            setattr(self, key, dict_[key])

    def parse_args(self, argv: Optional[List[str]]) -> Args:
        """
        Parse the arguments defined by args for the flitsr program with
        python's argparse. The result is an argparse Namespace object which
        includes all of the arguments parsed (or default values).

        Args:
          argv: The list of arguments to parse, usual taken from the command
                line arguments given
        """
        # Check file exists function
        def check_file(filename):
            if (osp.exists(filename)):
                return filename
            else:
                raise argparse.ArgumentTypeError('Could not find input file:'
                                                 f' \"{filename}\"')

        # check FLITSR type combinations function
        def check_type(type_comb: str):
            type_sep = type_comb.split('+')
            cluster = None
            ranker = None
            for t in type_sep:
                t = t.upper()
                if (hasattr(advanced.ClusterType, t)):
                    if (cluster is not None):
                        raise argparse.ArgumentTypeError('Cannot have two '
                            f'cluster types: {cluster.name} and {t}')
                    cluster = advanced.ClusterType[t]
                elif (hasattr(advanced.RankerType, t)):
                    if (ranker is not None):
                        raise argparse.ArgumentTypeError('Cannot have two '
                            f'ranker types: {ranker.name} and {t}')
                    ranker = advanced.RankerType[t]
                else:
                    raise argparse.ArgumentTypeError('Invalid type for '
                                                     f'--all-types: \"{t}\"')
            adv_type = Config(ranker, cluster)
            return adv_type

        # General options
        parser = argparse.ArgumentParser(prog='flitsr', description='An automatic '
                'fault finding/localization tool for multiple faults.')
        parser.add_argument('input', type=check_file,
                            help='The coverage file (TCM) or '
                            'directory (GZoltar) containing the coverage collected'
                            ' for the system over the test suite')
        parser.add_argument('-o', '--output', action='store', default=sys.stdout,
                type=argparse.FileType('w'), help='Specify the output file to use '
                'for all output (default: STDOUT).')
        parser.add_argument('--csv', action='store_true',
                help='By default FLITSR will output the ranking in it\'s own '
                'FLITSR ranking format. Enabling this option will allow FLITSR to '
                'output the ranking in CSV format compatible with GZoltar\'s CSV '
                'ranking format instead.')
        parser.add_argument('--spectrum-csv', action='store_true',
                help='Enabling this option will cause FLITSR to '
                'output the spectrum in CSV format.')
        default_metric = 'ochiai'
        parser.add_argument('-m', '--metric', dest='metrics', action='append',
                choices=Suspicious.getNames(True), metavar='METRIC',
                help='The underlying (SBFL) metric(s) to use when either ranking '
                '(if the sbfl option is given), or running the FLITSR algorithm. '
                'Option may be supplied multiple times to run multiple metrics. '
                'Specifying multiple metrics will output the results of each '
                'metric to a seperate file using the metric\'s name instead of '
                'stdout. Allowed values are: ['+', '.join(Suspicious.getNames(True))+'] '
                '(default: {})'.format(default_metric))
        parser.add_argument('-r', '--ranking', action='store_true',
                help='Changes FLITSR\'s expected input to be an SBFL ranking in '
                'Gzoltar or FLITSR format (determined automatically), instead of '
                'the usual coverage, and produces the specified calculations '
                '(or just the ranking if no calculations are given). NOTE: any '
                'non-output based options will be ignored with this option')
        parser.add_argument('-c', '--method', action='store_true',
                help='The default for FLITSR is to use the collected coverage '
                'as-is and merely produce the ranking in terms of the names/labels'
                ' given to the elements. Alternatively, using this option, FLITSR '
                'can assume the coverage given is a statement level coverage, and '
                'will attempt to collapse this coverage to produce a method level '
                'coverage result. This collapse is done by constructing a '
                'coverage matrix with only the method names, where the execution '
                'of a method is determined by the union of the executions of its '
                'statements. Bugs added to the coverage are handled in a similar '
                'fashion.')
        parser.add_argument('--split', help='When given, this option causes faults'
                ' that are a combination of two or more sub-faults in mutually'
                ' exclusive parts of the system to be split into separate'
                ' identified faults. As a by-product this also drops faults that'
                ' are not exposed by failing tests', action='store_true')
        parser.add_argument('--collapse', action='store_true',
                help='Collapse dynamic basic-block groups into singular elements '
                'for the ranking an calculations')
        parser.add_argument('-a', '--all', action='store_true',
                help='Used in the evaluation of FLITSR against other techniques. '
                'Runs all metrics given in suspicious.py and both FLITSR and '
                'FLITSR* extensions over each metric. Also enables all of the '
                'above evaluation calculations. Prints the results out to files '
                'named [<flitsr method>_]<metric>.run for each FLITSR method '
                'and metric')
        adv_types = list(advanced.all_types.keys())
        parser.add_argument('-t', '--types', action='append', type=check_type,
                             help='Specify the advanced type combination to use '
                             'when running FLITSR. Note that this argument '
                             'overrides any of the individual advanced type '
                             'arguments given to FLITSR (such as --multi, --sblf, etc.).'
                             'This argument may be specified multiple times to add '
                             'multiple type combinations to run. The format for '
                             'this argument is: "--types <type>[+<type>...]", '
                             'where each <type> is a (case-insensitive) FLITSR '
                             f'advanced type. Allowed types are: {adv_types}')
        parser.add_argument('--no-override', action='store_true',
                            help='By default FLITSR will override the output '
                            'file(s) if they already exist, printing a warning '
                            'message. This option instead allows FLITSR to leave '
                            'output files that already exist, skipping that '
                            'output and continuing with the rest of the outputs '
                            '(if any)')
        parser.add_argument('-d', '--decimals', action='store', type=int,
                default=2, help='Sets the precision (number of decimal points) for '
                'the output of all of the calculations (default: %(default)s)')

        # Advanced types options
        primitives = (bool, str, int, float, Path)
        adv_required_args: Dict[str, Tuple[str, List[str]]] = dict()

        advanced_groups: List[Tuple[str, Any, Any, Any]] = []

        refiner_enum = advanced.RefinerType
        if (len(refiner_enum) != 0):
            refiners = parser.add_argument_group('Spectrum refiner techniques',
                'One of the following spectrum refining techniques may be '
                'specified, along with it\'s options')
            refiner_mu = refiners.add_mutually_exclusive_group()
            advanced_groups.append(('refiner', refiner_enum, refiners,
                                    refiner_mu))

        cluster_enum = advanced.ClusterType
        if (len(cluster_enum) != 0):
            clusters = parser.add_argument_group('Clustering techniques',
                'One of the following clustering techniques may be specified, '
                'along with it\'s options')
            cluster_mu = clusters.add_mutually_exclusive_group()
            advanced_groups.append(('cluster', cluster_enum, clusters,
                                    cluster_mu))

        ranker_enum = advanced.RankerType
        if (len(ranker_enum) != 0):
            rankers = parser.add_argument_group('Ranking techniques',
                'One of the following advanced ranking techniques may be '
                'specified, along with it\'s options (default: flitsr)')
            ranker_mu = rankers.add_mutually_exclusive_group()
            advanced_groups.append(('ranker', ranker_enum, rankers, ranker_mu))

        for adv_name, adv_enum, group, mu in advanced_groups:
            for type_ in list(adv_enum):
                name, class_ = type_.name, type_.value
                init = class_.__init__
                disp_name = name.lower()
                help_ = ''  # '(default: %(default)s)'
                # add docstring
                if (class_.__doc__ is not None and class_.__doc__ != ''):
                    help_ = class_.__doc__ + ' ' + help_
                # add cmd line argument for this advanced type
                mu.add_argument('--'+disp_name, dest=adv_name,
                                action='store_const', const=adv_enum[name],
                                help=help_)
                argspec = inspect.getfullargspec(init)
                def_diff = (len(argspec.args)-1) - (0 if
                                                    argspec.defaults is None
                                                    else len(argspec.defaults))
                # add cmd-line arguments for each parameter of this adv type
                self._advanced_params[name] = list()
                for p_index, param in enumerate(argspec.args[1:]):
                    self._advanced_params[name].append(param)
                    # skip adding this parameter if it is marked as existing
                    if (hasattr(init, '__existing__') and
                        param in init.__existing__):
                        continue
                    paramName = '--'+disp_name+'-'+param.replace('_', '-')
                    parser_args: Dict[str, Any] = {}
                    # add default arguments if param has
                    if (argspec.defaults is not None and p_index >= def_diff):
                        parser_args['default'] = argspec.defaults[p_index-def_diff]
                        parser_args['help'] = '(default: %(default)s)'
                    # else add param as conditionally "required" if not bool
                    elif (argspec.annotations[param] is not bool):
                        if (name not in adv_required_args):
                            adv_required_args[name] = (adv_name, list())
                        adv_required_args[name][1].append(param)
                        parser_args['help'] = '(required)'
                    # add choices
                    if (hasattr(init, '__choices__') and
                        param in init.__choices__):
                        parser_args['choices'] = init.__choices__[param]
                    # add types
                    if (param in argspec.annotations):
                        paramType = argspec.annotations[param]
                        # specially handle bools
                        if (paramType is bool):
                            if ('default' in parser_args):
                                # check if default is True
                                if (parser_args['default']):
                                    parser_args['dest'] = disp_name+'_'+param
                                    paramName = ('--' + disp_name + '-no-' +
                                                 param.replace('_', '-'))
                                    parser_args['action'] = 'store_false'
                                else:
                                    parser_args['action'] = 'store_true'
                            else:
                                parser_args['action'] = 'store_true'
                        # specially handle file IO types
                        elif (issubclass(paramType, IO)):
                            if (issubclass(paramType, BinaryIO)):
                                parser_args['type'] = argparse.FileType('rb')
                            else:
                                parser_args['type'] = \
                                  argparse.FileType('r', encoding='UTF-8')
                        else:
                            if (paramType not in primitives):
                                if (not hasattr(class_, f'_{param}')):
                                    err = ('Could not find type conversion '
                                           f'for {param} in {name}')
                                    raise NameError(err)
                                paramType = getattr(class_, f'_{param}')
                            parser_args['type'] = paramType
                            if ('choices' not in parser_args):
                                parser_args['metavar'] = param
                    # finalize option
                    group.add_argument(paramName, **parser_args)

        # manually set flitsr as the default
        parser.set_defaults(ranker=advanced.RankerType['FLITSR'])

        # parser.add_argument('--multi', action='store_true',
        #         help='Runs the FLITSR* (i.e. multi-round) algorithm')
        # parser.add_argument('-i', '--internal-ranking', action='store',
        #         choices=['flitsr', 'reverse', 'original', 'auto', 'conf'], default='auto',
        #         help='Specify the order in which the elements of each FLITSR basis '
        #         'are ranked. "flitsr" uses the order that FLITSR returns the basis '
        #         'in (i.e. from FLITSRs lowest to highest recursion depth), which '
        #         'aligns with FLITSRs confidence for each element being a fault. '
        #         '"reverse" uses the reverse of "flitsr" (i.e. the order in which '
        #         'FLITSR identifies the elements) which gives elements that use a '
        #         'larger part of the original test suite first. "original" returns '
        #         'the elements based on their original positions in the ranking '
        #         'produced by the base SBFL metric used by FLITSR.')

        # Tie breaking options
        tie_grp = parser.add_argument_group('Tie breaking strategy',
                                            'Specifies the tie breaking '
                                            'strategy to use for FLITSR and '
                                            'the localization')
        tie_grp.add_argument('--tiebrk', action='store_const',
                             default=Tiebrk.ORIG, const=Tiebrk.EXEC,
                             help='Breaks ties using only execution counts')
        tie_grp.add_argument('--rndm', action='store_const', const=Tiebrk.RNDM,
                             dest='tiebrk', help='Breaks ties by a randomly '
                             'ordering')
        tie_grp.add_argument('--otie', action='store_const', const=Tiebrk.ORIG,
                             dest='tiebrk', help='Breaks ties by using the '
                             'original base metric ranking (in the case of '
                             'FLITSR) and by execution counts otherwise')

        # Calculation options
        calc_grp = parser.add_argument_group('Calculations',
                'The following arguments replace the default ranking output of '
                'FLITSR with evaluation calculations. Multiple of the following '
                'arguments can be given in the same call')

        # check type of general weffort
        def check_fault_type(nth):
            if (int(nth) > 0):
                return int(nth)
            else:
                raise argparse.ArgumentTypeError(f'Invalid fault number {nth}')

        # Wasted effort calcuation options
        calc_grp.add_argument('--first', dest='weff', action='append_const',
                default=[], const='first',
                help='Display the wasted effort to the first fault')
        calc_grp.add_argument('--avg', '--average', dest='weff',
                action='append_const', const='avg',
                help='Display the wasted effort to the average fault')
        calc_grp.add_argument('--med', '--median', dest='weff',
                action='append_const', const='med',
                help='Display the wasted effort to the median fault')
        calc_grp.add_argument('--last', dest='weff', action='append_const',
                const='last', help='Display the wasted effort to the last fault')
        calc_grp.add_argument('--weffort', dest='weff', action='append', metavar='N',
                type=check_fault_type, help='Display the wasted effort to the Nth fault')

        # TOP1 calculation options
        calc_grp.add_argument('--one-top1', dest='top1',
                action='append_const', const='one', default=[],
                help='Display a boolean value indicating whether at least one '
                'fault was found in the TOP1 group (elements with the highest '
                'suspiciousness')
        calc_grp.add_argument('--all-top1', dest='top1',
                action='append_const', const='all',
                help='Display the number of faults found in the top1 group')
        calc_grp.add_argument('--perc-top1', dest='top1',
                action='append_const', const='perc',
                help='Display the percentage of faults found in the top1 group')

        # Percent at n calculation options
        calc_grp.add_argument('--perc@n', '--percent-at-n', dest='perc_at_n',
                action='append_const', const='perc', default=[],
                help='Produces the percentage-at-N values (i.e. the percentage of '
                'faults found at N%% of code inspected). The output of the perc@n '
                'calculation is a list of ranks of all found faults, preceded by '
                'the number of elements in the system, which can be used to '
                'generate percentage-at-N/recall graphs')
        calc_grp.add_argument('--auc', '--area-under-curve', dest='perc_at_n',
                action='append_const', const='auc',
                help='Dislpays the area under the curve produced by the '
                'percentage-at-N calculation')
        calc_grp.add_argument('--pauc', '--percent-area-under-curve',
                dest='perc_at_n', action='append_const', const='pauc',
                help='Dislpays the area under the curve '
                'produced by the percentage-at-N calculation as a percentage of '
                'the maximum possible value (i.e. closest to perfect recall)')
        calc_grp.add_argument('--lauc', '--log-area-under-curve',
                dest='perc_at_n', action='append_const', const='lauc',
                help='Dislpays the area under the curve '
                'produced by the percentage-at-N calculation as a logarithmic '
                'percentage of the maximum possible value (i.e. closest to perfect '
                'recall). The logarithmic effect causes lower ranks to have a '
                'greater effect on the value, which corresponds to the lower ranks '
                'being more useful to the developer')

        # Precision recall type functions
        def pr(value):
            if (value == "b" or value == "f"):
                return value
            elif (value.isdigit()):
                return int(value)
            else:
                raise ValueError(str(value)+" is not a valid precision/recall "
                        "string value")
        def precision(value):
            return ('p', pr(value))
        def recall(value):
            return ('r', pr(value))

        # Precision recall calculation options
        calc_grp.add_argument('--precision-at', dest='prec_rec', metavar='x',
                action='append', default=[], type=precision,
                help='Displays precision values at a given rank `x`. Precision '
                'is the amount of faults f found within the cut-off point `x`, '
                'out of the number of elements seen (i.e. f/x). Can be specified '
                'multiple times')
        calc_grp.add_argument('--recall-at', dest='prec_rec', metavar='x',
                action='append', default=[], type=recall,
                help='Displays recall values at a given rank `x`. Recall '
                'is the amount of faults f found within the cut-off point `x`, '
                'out of the total number of faults n (i.e. f/n). Can be specified '
                'multiple times')

        # Fault output options
        calc_grp.add_argument('--fault-num', dest='faults', action='append_const',
                              const='num', default=[],
                              help='Display the number of faults in the program')
        calc_grp.add_argument('--fault-ids', dest='faults', action='append_const',
                              const='ids',
                              help='Display the IDs of the faults in the program')
        calc_grp.add_argument('--fault-elems', dest='faults',
                              action='append_const', const='elems',
                              help='Display the elements that are faulty in the program')
        calc_grp.add_argument('--fault-all', dest='faults', action='append_const',
                              const='all',
                              help='Display all info of the faults in the program')

        # parser.add_argument('-p', '--parallel', action='store',
        #          choices=parallel_opts, metavar='ALGORITHM',
        #          help='Run one of the parallel debugging algorithms on the spectrum '
        #          'to produce multiple spectrums, and process all other options on '
        #          'each spectrum. Allowed values are: ['+', '.join(parallel_opts)+']')

        # parser.add_argument('--artemis', action='store_true',
        #                     help='Run the ARTEMIS technique on the spectrum to '
        #                     'produce the ranked lists. This option may be '
        #                     'combined with FLITSR and parallel to produce a '
        #                     'hybrid technique.')

        # TODO: Add OBA and worst args if necessary
        cut_eval_opts=['worst', 'best', 'resolve']
        parser.add_argument('--cutoff-eval', action='store', metavar='MODE',
                choices=cut_eval_opts, help='Specifies the performance mode to use '
                'when using a multi-fault fixing cut-off strategy to produce '
                'rankings. Allowed values are: ['+', '.join(cut_eval_opts)+']'+
                ' (default %(default)s)', default='worst')
        cutoff_opts = cutoff_points.getNames()
        parser.add_argument('--cutoff-strategy', action='store', metavar='STRATEGY',
                help='Cuts off the ranking using the given '
                'strategy\'s cut-off point. This affects both the rank output '
                'method and any calculations. Allowed values are: ['+
                ', '.join(cutoff_opts)+'] (default %(default)s). '
                'For basis, an optional value n may be given (e.g. basis=n) '
                'that determines the number of bases included before the cutoff')

        argcomplete.autocomplete(parser)

        args = parser.parse_args(argv)
        # check "required" advanced type arguments
        for adv_name, (adv_type, adv_args) in adv_required_args.items():
            if (getattr(args, adv_type) is not None and
                getattr(args, adv_type).name == adv_name):
                dname = adv_name.lower()
                for adv_arg in adv_args:
                    if (getattr(args, dname+'_'+adv_arg) is None):
                        err = (f'--{dname}-{adv_arg} is required when '
                               f'--{dname} is used')
                        parser.error(err)
        # Set the metrics based on 'all' or the default metric
        if (args.metrics is None):
            if (args.all is True):
                args.metrics = Suspicious.getNames()
            else:
                args.metrics = [default_metric]
        # Set the flitsr types based on 'all' or what is set
        if (args.all is True):
            if (args.types is None):
                ts = {}
                if (getattr(args, 'refiner', None) is not None):
                    ts['refiner'] = args.refiner
                if (getattr(args, 'cluster', None) is not None):
                    ts['cluster'] = args.cluster
                args.types = [Config(**ts),
                              Config(advanced.RankerType['FLITSR'], **ts),
                              Config(advanced.RankerType['MULTI'], **ts)]
            if (len(args.weff) == 0 and len(args.top1) == 0 and
                    len(args.perc_at_n) == 0 and len(args.prec_rec) == 0):
                args.weff = ["first", "avg", "med", "last", 2, 3, 5]
                args.perc_at_n = ["perc"]
                args.prec_rec = [('p', 1), ('p', 5), ('p', 10), ('p', "f"),
                                 ('r', 1), ('r', 5), ('r', 10), ('r', "f")]
                args.faults = ["num"]
        elif (args.types is None):
            args.types = [Config(args.ranker, getattr(args, 'cluster', None),
                                 getattr(args, 'refiner', None))]
        if (args.cutoff_strategy and args.cutoff_strategy.startswith('basis')):
            args.sbfl = False
            args.multi = 1
        self._add_params(args)
        return self

    def get_arg_group(self, group_name):
        group = {}
        params = self._advanced_params[group_name.upper()]
        prefix = group_name.lower()+'_'
        for param in params:
            if (hasattr(self, prefix+param)):
                group[param] = getattr(self, prefix+param)
            elif (hasattr(self, param)):
                group[param] = getattr(self, param)
            elif (param == 'args'):  # special case for full args
                group[param] = self
            else:
                raise KeyError(f'Could not find {param} in args!')
        return group


if __name__ == "__main__":
    args = Args().parse_args(sys.argv[1:])
    print(args)



from typing import List
from abc import ABC, abstractmethod
from flitsr.spectrum import Spectrum
from flitsr import input


class Input(ABC):
    def __init_subclass__(cls):
        input.register_input(cls)

    @abstractmethod
    def get_run_file_name(self, input_path: str):
        pass

    @abstractmethod
    def read_spectrum(self, input_path: str, split_faults: bool,
                      method_level=False) -> Spectrum:
        pass

    @classmethod
    def write_spectrum(cls, spectrum: Spectrum, output_path: str):
        raise NotImplementedError("Cannot write spectrum in"
                                  f"{cls.__name__} format")

    @classmethod
    def get_elem_separators(cls) -> List[str]:
        raise NotImplementedError(f"Input type {cls.__name__} not supported "
                                  "for output string")

    @classmethod
    def get_type(cls) -> 'input.InputType':
        return input.InputType[cls.__name__.upper()]

    @staticmethod
    @abstractmethod
    def check_format(input_path: str) -> bool:
        """
        Check whether the files from the specified input_path are of the
        concrete types format. This implementation of this abstract method
        must return a boolean, where True indicates that the input is in the
        format recognized by the implementing class, and False otherwise
        """
        pass

    @staticmethod
    def read_in(input_path: str, split_faults: bool,
                method_level=False) -> Spectrum:
        reader = Input.get_reader(input_path)
        return reader.read_spectrum(input_path, split_faults,
                                    method_level)

    @staticmethod
    def get_reader(input_path: str) -> 'Input':
        for input_enum in list(input.InputType):
            input_cls = input_enum.value
            if (input_cls.check_format(input_path)):
                return input_cls()
        raise ValueError(f"Unknown input type \"{input_path}\"")



import sys
from matplotlib import pyplot as plt # type: ignore


def distro(file):
    lines = open(file).readlines()
    x = []
    y = []
    for line in lines:
        line = line.strip()
        if (line):
            vals = line.split()
            x.append(int(vals[0]))
            y.append(int(vals[1]))
    return x, y


if __name__ == "__main__":
    x, y = distro(sys.argv[1])
    plt.bar(x, y)
    plt.title("Fault distribution for project "+str(sys.argv[2]))
    plt.xlim([0, 32])
    plt.ylim([0, 1200])
    plt.show()
